/*
 * antibioticconcentration.cpp
 *
 * Created by epienaar on 06/15/2015.
 */


#include "antibioticconcentration.h"

// Deep copy an antibiotic concentration list.

// Note that as part of cloning from one GrSimulation object to another, the
// AntibioticSimulationGrid pointers in each AntibioticConcentration object
// cannot be updated here, since we don't have access to them.  They will need
// to be updated later by a higher level function that does have access to
// them, by calling function  AntibioticConcentration::updateAntibioticConcentrationList.
void AntibioticConcentration::cloneAntibioticConcentrationList(
         const AntibioticConcentrationList& srcList,
         AntibioticConcentrationList& dstList)
{
    assert(&dstList != &srcList);

    AntibioticConcentrationList::const_iterator srcIt = srcList.begin();
    AntibioticConcentrationList::iterator dstIt = dstList.begin();
    for (; srcIt != srcList.end(); ++srcIt, ++dstIt)
    {
        (*dstIt) = new AntibioticConcentration(**srcIt);
    }
}

// Update an antibiotic concentration list with pointers to a clone
// GrSimulation object's AntibioticSimulationGrids.
void AntibioticConcentration::updateAntibioticConcentrationList(
	AntibioticConcentrationList& cloneAbxConcentrationList,
        const AntibioticSimulationGridList& cloneAbxGridList,
        string& errorText)
{
	if(cloneAbxConcentrationList.size() != cloneAbxGridList.size())
	{
	  cerr << "GrSimulation::clone:" << endl
		   << "The antibiotic concentration list for " << errorText
		   << " has size " << cloneAbxConcentrationList.size()
		   << ", which does not match the antibiotic simulation grid list size of "
		   << cloneAbxGridList.size() << "."
		   << endl;
		   throw std::runtime_error("Fatal error");
	}

	AntibioticConcentrationList::iterator abxConcIt = cloneAbxConcentrationList.begin();
	AntibioticSimulationGridList::const_iterator abxSimGridIt = cloneAbxGridList.begin();

	for(; abxConcIt != cloneAbxConcentrationList.end(); ++abxConcIt, ++abxSimGridIt)
	{
	  (*abxConcIt)->setAntibioticSimulationGrid(*abxSimGridIt);
	}

}

AntibioticConcentration::AntibioticConcentration()
    : _abxgrid()
    , _concentration(0.0)
    , _isGeneticallyResistant(false)
    , _hasCompensatoryMutation(false)
    , _mtbSpecificC50Extra(0.0)
    , _mtbSpecificC50Intra(0.0)
    , _mtbSpecificC50Nonrep(0.0)
    , _hasEffluxPumps(false)
    , _timeInEffluxInducingRange()

{
}

AntibioticConcentration::AntibioticConcentration(AntibioticSimulationGrid* antibioticSimulationGrid)
    : _abxgrid(antibioticSimulationGrid)
    , _concentration(0.0)
    , _isGeneticallyResistant(false)
    , _hasCompensatoryMutation(false)
    , _mtbSpecificC50Extra(antibioticSimulationGrid->c50Extra())
    , _mtbSpecificC50Intra(antibioticSimulationGrid->c50Intra())
    , _mtbSpecificC50Nonrep(antibioticSimulationGrid->c50Nonrep())
    , _hasEffluxPumps(false)
    , _timeInEffluxInducingRange(_PARAM(Drugs_effluxInductionStatInterval), false)

{
//    cout << "isResistant: " << _isGeneticallyResistant << endl;
}


AntibioticConcentration::~AntibioticConcentration()
{
}


// This function is called when a saved state is loaded. The drug parameters
// might be different than they were when the state was saved, so the C50
// values will need to be udpated, to be based on the new drug parameter.

void AntibioticConcentration::updateMtbSpecificC50()
{
    _mtbSpecificC50Extra = _abxgrid->c50Extra();
    _mtbSpecificC50Intra = _abxgrid->c50Intra();
    _mtbSpecificC50Nonrep = _abxgrid->c50Nonrep();
    
    if(_isGeneticallyResistant)
    {
        changeC50ForResistance();
    }
}

// Change the C50 to reflect the resistance
void AntibioticConcentration::changeC50ForResistance()
{
    double resistanceC50increase = _abxgrid->foldIncreaseC50FromResistance();

    _mtbSpecificC50Extra *= resistanceC50increase;
    _mtbSpecificC50Intra *= resistanceC50increase;
    _mtbSpecificC50Nonrep *= resistanceC50increase;
}

double AntibioticConcentration::effectiveConcentration(double conc)
{
    AntibioticSimulationGrid* asg = antibioticSimulationGrid();
    // if this bug has efflux pumps turned on for this drug then change the concentration according to the param
    if (hasEffluxPumps())
    {
        conc = conc * asg->concentrationChangeByEfflux();
    }
    return conc;
}

double AntibioticConcentration::killingRateConstFromEmax(double emax, double effectiveConc, double hill, double c50)
{
    double killingRateConst = emax * (pow(effectiveConc,hill) / ( pow(c50,hill) + pow(effectiveConc,hill) ));
    return killingRateConst;
}



