/*
* antibiotic.h
*
*
* Created by epienaar on 03/30/15.
*/


#ifndef ANTIBIOTICSIMULATIONGRID_H
#define ANTIBIOTICSIMULATIONGRID_H

#include "grid-selector.h"
#include "gr.h"
#include "pos-selector.h"
#include "lungparams.h"
#include "grdiffusionFFT.h"
#include "grdiffusion-crank-nicolson-adi.h"
#include "graccumulator.h"
#include "numericalMethods.h"
#include "drug-kill.h"

#include <iostream>
#include <string>
#include <vector>

using std::valarray;
struct GridAntibioticPartitioningFunc;

using namespace std;

// =========================================
// AntibioticGrid class
// Make a class for antibiotic grids to be able to track concentration and u and v grids
// for each antibiotic in a currentGrid and nextGrid to be able to diffuse.

class AntibioticGrid
{

private:
    // Concentration grids
    std::vector<double> _grid;            // free abx grid - diffuses
    std::vector<double> _u_grid;
    std::vector<double> _v_grid;
    std::vector<double> _casBnd_grid;     // caseum bound grid - does not diffuse
    std::vector<double> _epithBnd_grid;     // epithelium bound grid - does not diffuse
    Pos _dim;

public:
	const Pos& getRange() const;
	int getSize() const;
	const Pos getCenter() const;
    const double& grid(const Pos& p) const;
    void setGrid(const Pos& p, double v);
    void incGrid(const Pos& p, double v);
	void setEntireGrid(double value);
    const double* grid() const;
    double* grid();
    const std::vector<double>& gridVector() const;
    std::vector<double>& gridVector();

    const double& u_grid(const Pos& p) const;
    const double* u_grid() const;
    double* u_grid();

    const double& v_grid(const Pos& p) const;
    const double* v_grid() const;
    double* v_grid();

    const double& casBnd_grid(const Pos& p) const;
    void setCasBnd_grid(const Pos& p, double v);
    void incCasBnd_grid(const Pos& p, double v);
    const double* casBnd_grid() const;
    double* casBnd_grid();
    const std::vector<double>& casBndGridVector() const;
    std::vector<double>& casBndGridVector();

    const double& epithBnd_grid(const Pos& p) const;
    void setEpithBnd_grid(const Pos& p, double v);
    void incEpithBnd_grid(const Pos& p, double v);
    const double* epithBnd_grid() const;
    double* epithBnd_grid();
    const std::vector<double>& epithBndGridVector() const;
    std::vector<double>& epithBndGridVector();

    void displayGrid(double* grid, int floatPrecision, bool isPadded);

    AntibioticGrid();
    AntibioticGrid(const Pos &dim);
    AntibioticGrid(const AntibioticGrid& ag);
    virtual ~AntibioticGrid();

private:
    friend class boost::serialization::access; 
    template<class Archive>
    void serialize(Archive& ar, const unsigned int version);

};

inline const Pos& AntibioticGrid::getRange() const
{
	return _dim;
}

inline int AntibioticGrid::getSize() const
{
	return _dim.size();
}

inline const Pos AntibioticGrid::getCenter() const
{
	return _dim / 2;
}

inline const double& AntibioticGrid::grid(const Pos& p) const
{
    return _grid[GridSpec::padInd(_dim, p)];
}

inline void AntibioticGrid::setGrid(const Pos& p, double v)
{
    _u_grid[GridSpec::padInd(_dim, p)] += v - _grid[GridSpec::padInd(_dim, p)];
    _v_grid[GridSpec::padInd(_dim, p)] += v - _grid[GridSpec::padInd(_dim, p)];
    _grid[GridSpec::padInd(_dim, p)] = v;
}
inline void AntibioticGrid::incGrid(const Pos& p, double v)
{
    _u_grid[GridSpec::padInd(_dim, p)] += v;
    _v_grid[GridSpec::padInd(_dim, p)] += v;
    _grid[GridSpec::padInd(_dim, p)] += v;
}

inline void AntibioticGrid::setEntireGrid(double value)
{
    GridSpec::EntireGridIterator egi = GridSpec::egiBegin(_dim);
    GridSpec::EntireGridIterator egiEnd =  GridSpec::egiEnd(_dim);

    for(; egi != egiEnd; ++egi)
    {
	  const Pos& pos = *egi;
      setGrid(pos, value);
    }
}

inline const double* AntibioticGrid::grid() const
{
    return _grid.data();
}
inline double* AntibioticGrid::grid()
{
    return _grid.data();
}

inline const std::vector<double>& AntibioticGrid::gridVector() const
{
    return _grid;
}

inline std::vector<double>& AntibioticGrid::gridVector()
{
    return _grid;
}

inline const double* AntibioticGrid::u_grid() const
{
    return _u_grid.data();
}
inline const double& AntibioticGrid::u_grid(const Pos& p) const
{
    return _u_grid[GridSpec::padInd(_dim, p)];
}
inline double* AntibioticGrid::u_grid()
{
    return _u_grid.data();
}
inline const double& AntibioticGrid::v_grid(const Pos& p) const
{
    return _v_grid[GridSpec::padInd(_dim, p)];
}
inline const double* AntibioticGrid::v_grid() const
{
    return _v_grid.data();
}
inline double* AntibioticGrid::v_grid()
{
    return _v_grid.data();
}

inline const double& AntibioticGrid::casBnd_grid(const Pos& p) const
{
    return _casBnd_grid[GridSpec::ind(_dim, p)];
}

inline void AntibioticGrid::setCasBnd_grid(const Pos& p, double v)
{
    _casBnd_grid[GridSpec::ind(_dim, p)] = v;
}
inline void AntibioticGrid::incCasBnd_grid(const Pos& p, double v)
{
    _casBnd_grid[GridSpec::ind(_dim, p)] += v;
}
inline const double* AntibioticGrid::casBnd_grid() const
{
    return _casBnd_grid.data();
}
inline double* AntibioticGrid::casBnd_grid()
{
    return _casBnd_grid.data();
}
inline const std::vector<double>& AntibioticGrid::casBndGridVector() const
{
    return _casBnd_grid;
}

inline std::vector<double>& AntibioticGrid::casBndGridVector()
{
    return _casBnd_grid;
}
inline const double& AntibioticGrid::epithBnd_grid(const Pos& p) const
{
    return _epithBnd_grid[GridSpec::ind(_dim, p)];
}
inline void AntibioticGrid::setEpithBnd_grid(const Pos& p, double v)
{
    _epithBnd_grid[GridSpec::ind(_dim, p)] = v;
}
inline void AntibioticGrid::incEpithBnd_grid(const Pos& p, double v)
{
    _epithBnd_grid[GridSpec::ind(_dim, p)] += v;
}
inline const double* AntibioticGrid::epithBnd_grid() const
{
    return _epithBnd_grid.data();
}
inline double* AntibioticGrid::epithBnd_grid()
{
    return _epithBnd_grid.data();
}
inline const std::vector<double>& AntibioticGrid::epithBndGridVector() const
{
    return _epithBnd_grid;
}

inline std::vector<double>& AntibioticGrid::epithBndGridVector()
{
    return _epithBnd_grid;
}


inline void AntibioticGrid::displayGrid(double* grid, int floatPrecision, bool isPadded)
{
    int oldFloatPrecision = cerr.precision();
    cerr.precision(floatPrecision);
    cerr << std::fixed;
    cerr << endl << "ABX" << ": ";

    double total = 0;

    GridSpec::EntireGridIteratorFormatting egif = GridSpec::egifBegin(_dim);
    GridSpec::EntireGridIteratorFormatting egifEnd =  GridSpec::egifEnd(_dim);

    for(; egif != egifEnd; ++egif)
    {
      GridSpec::egifRetval egifr = *egif;
      const Pos& p = egifr.first;
      int dimensionEndIndicator = egifr.second;

      if(dimensionEndIndicator < _dim.x)
      {
        cerr << endl << "x:" << p.x;
      }

      double value;
      if (isPadded)
      {
          value = grid[GridSpec::padInd(_dim, p)];
      }
      else
      {
          value = grid[GridSpec::ind(_dim, p)];
      }

      total += value;
      cerr << "\t" << value;
    }
    cerr << endl; // End of last row.
    cerr << "Antibiotic total: " << total << endl << endl;
    cerr.precision(oldFloatPrecision);
}


template<class Archive>
void AntibioticGrid::serialize(Archive& ar, const unsigned int /*version*/)
{
    // Concentration Grids
    ar & BOOST_SERIALIZATION_NVP(_grid);
    ar & BOOST_SERIALIZATION_NVP(_u_grid);
    ar & BOOST_SERIALIZATION_NVP(_v_grid);
    ar & BOOST_SERIALIZATION_NVP(_casBnd_grid);
    ar & BOOST_SERIALIZATION_NVP(_epithBnd_grid);
    ar & BOOST_SERIALIZATION_NVP(_dim);
}

// END AntibioticGrid class
// =========================================



// =========================================
// DrugMtbStat class
// Make a class for bug antibiotic stats that have different values for each
// bug subpopulation (intracellular, extracellular, nonrep, total)
class DrugMtbStat
{
public:
    int _extMtb;
    int _nonrepExtMtb;
    int _intMtb;
    int _totMtb;

    void reset();

    DrugMtbStat();
    virtual ~DrugMtbStat();

private:
    friend class boost::serialization::access; 
    template<class Archive>
    void serialize(Archive& ar, const unsigned int version);
};


inline void DrugMtbStat::reset()
{
    _extMtb = 0;
    _nonrepExtMtb = 0;
    _intMtb = 0;
    _totMtb = 0;
}

template<class Archive>
void DrugMtbStat::serialize(Archive& ar, const unsigned int /*version*/)
{
    // Concentration Grids
    ar & BOOST_SERIALIZATION_NVP(_extMtb);
    ar & BOOST_SERIALIZATION_NVP(_nonrepExtMtb);
    ar & BOOST_SERIALIZATION_NVP(_intMtb);
    ar & BOOST_SERIALIZATION_NVP(_totMtb);
}

// END DrugMtbStat
// =========================================



class AntibioticSimulationGrid
{
public:

    const static std::string antibioticStatHeaders[];
    const static int nrHeaders;

    const static std::string antibioticPlasmaParamHeaders[];
    const static int nrPlasmaParamHeaders;

    static double computeInteractEffectiveConc(double fic, const vector<double>& concRatio);
    static double computeInteractAdjConc(double fic, const vector<double>& concVec, const vector<double>& emaxVec,
                                         const vector<double>& c50Vec, const vector<double>& hVec,
                                         double intc50, double intEmax, double intH);
    static bool checkShouldKill(double killRate);

private:

    Pos _dim;

    // abbreviation of the antibiotic
    string _name;

    // cutoff value for drug concentrations. Any drug concentration below this
    // cutoff is set to 0.
    double _cutOffValue;

    // Parameters
    /*
     * !!! If the parameter members change then the serialize and deserialize functions need to be updated AS WELL AS the loadParameters function !!!
     */
    int _PKModel;
    double _vesselActivity;
    int _partitionUsingSteadyState;
    double _degRateConst;
    double _degRateConstInt;
    double _diffusivity;
    double _minDiffusivity;
    double _kAbs;
    double _q;
    double _volDistCentral;     // This is assuming a 2 compartment model as for INH and RIF - will need to be modified if other drugs have different PK structures
    double _volDistPeriph;      // This is assuming a 2 compartment model as for INH and RIF - will need to be modified if other drugs have different PK structures
    double _clearance;
    double _dose;
    int _dosageInterval;
    double _cellUptake;
    double _vascularPermeability;
    double _permCoeff;
    double _intraActivity;
    double _extraActivity;
    double _nonrepActivity;
    double _c50Extra;
    double _c50Intra;
    double _c50Nonrep;
    double _consumptionRate;
    double _consumptionHalfSat;
    double _mic;
    double _hillExtra;
    double _hillIntra;
    double _hillNonrep;
    double _caseumUnboundFraction;
    double _caseumBindingRateConst;
    double _epitheliumBindingKa;
    double _epitheliumBindingRateConst;
    double _cellularKout;
    double _MW;
    int _resistRiskIndicatorFlag;
    double _aucRiskLB;
    double _aucRiskUB;
    double _cmaxRiskLB;
    double _cmaxRiskUB;
    double _numberResistMutations;
    double _relativeFitnessResist;
    double _numberCompensatoryMutations;
    double _foldIncreaseC50FromResistance;
    int _numTimeStepsEffluxInduction;
    double _minEffluxInduction;         /* Minimum concentration where efflux pumps are induced (as a fraction of MIC).*/
    double _maxEffluxInduction;         /* Maximum concentration where efflux pumps are induced (as a fraction of MIC).*/
    double _concentrationChangeByEfflux;
    int _numberOfRandomMonoResistantBugsToAdd;
    double _volDistLung;

    // Concentration Grids
    std::vector<double> _pMat;

    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    CrankNicolsonADIChemical* _cnadiChemical;
    #endif

    AntibioticGrid* _currentGrid;
    AntibioticGrid* _nextGrid;

    // Vascular variables
    double _bloodConcentration;        // concentration in blood/central compartment (mol/L)
    double _absComp1Concentration;     // concentration in 1st absorption transit compartment (mg/kg)
    double _absComp2Concentration;     // concentration in 2nd absorption transit compartment (mg/kg)
    double _perCompConcentration;      // concentration in peripheral distribution compartment (mg/kg)
    double _lungCompConcentration;     // concentration in lung compartment (mg/kg)
    double _intMacConcentration;       // concentration in lung macs (mg/L)

    std::valarray<double> _initvector;  // valarray that holds all the integration variables to be passed to an integration method

    // Stats Grids
    std::vector<double> _intracellularGrid;
    std::vector<double> _totalGrid;
    std::vector<double> _aucIntraGrid;
    std::vector<double> _aucExtraGrid;
    std::vector<double> _aucTotalGrid;
    std::vector<double> _cmaxIntraGrid;
    std::vector<double> _cmaxExtraGrid;
    std::vector<double> _cmaxTotalGrid;

    // Stats
    double _totalWholeGrid;                         // mg/L
    double _totalGran;                              // mg/L
    double _totalNorm;                              // mg/L
    double _totalCaseum;                            // mg/L
    double _partialCaseum;                            // mg/L
    double _avgConcInfInt;                          // mg/L
    double _avgConcExtMtb;                          // mg/L
    double _avgConcNonMtb;                          // mg/L

    double _drugConcentrationArea;                  // "Area Under Drug Concentration Curve" (mg.h/L)
    double _drugConcentrationAreaNorm;              // "Area Under Drug Concentration Curve outside granuloma" (mg.h/L)
    double _drugConcentrationAreaCaseum;            // "Area Under Drug Concentration Curve in caseum" (mg.h/L)
    double _drugConcentrationAreaGran;              // "Area Under Drug Concentration Curve inside granuloma" (mg.h/L)
    double _drugConcentrationAreaBlood;             // "Area Under Drug Concentration Curve in blood" (mg.h/L)
    double _drugConcentrationAreaLung;              // "Area Under Drug Concentration Curve in Lung" (mg.h/L)
    double _drugConcentrationAreaPeriph;            // "Area Under Drug Concentration Curve in Lung" (mg.h/L)
    double _cmax;                                   // "Cmax for the period where AUC is measured for whole grid" (mg/L)
    double _cmaxNorm;                               // "Cmax for the period where AUC is measured in 'normal' compartments" (mg/L)
    double _cmaxCaseum;                             // "Cmax for the period where AUC is measured in caseated compartments" (mg/L)
    double _cmaxGran;                               // "Cmax for the period where AUC is measured in granuloma compartments" (mg/L)
    double _cmaxBlood;                              // "Cmax for the period where AUC is measured in blood" (mg/L)
    double _compOverMIC;                            // "Number of compartments with concentration over MIC"
    double _extMtbOverMIC;                          // "Number of extracellular bacteria in compartments with concentration over MIC"
    double _macsOverMIC;                            // "Number of macrophages with intracellular concentration over MIC"
    double _intMtbOverMIC;                          // "Number of intracellular bacteria in macrophages with concentration over MIC"
    double _intMtbHighRiskOfResistance;             // "Number of intracellular bacteria in exposure with high risk of resistance "
    double _extMtbHighRiskOfResistance;             // "Number of extracellular bacteria in exposure with high risk of resistance "
    double _nonrepMtbHighRiskOfResistance;          // "Number of total bacteria in exposure with high risk of resistance "
    double _macsHighRiskOfResistance;               // "Number macrophages where intracellular bacteria are in exposure with high risk of resistance "
    double _compHighRiskOfResistance;               // "Number grid compartments where extracellular bacteria are in exposure with high risk of resistance "
    double _intMtbSingleDrugKill;                   // "Number of intracellular mtb that are killed by the single antibiotic specific to this grid"
    double _extMtbSingleDrugKill;                   // "Number of extracellular mtb that are killed by the single antibiotic specific to this grid"
    double _nonrepSingleDrugKill;                   // "Number of non-replicating extracellular mtb that are killed by the single antibiotic specific to this grid"

    // Resistance stats
    DrugMtbStat _resistant;                         // "Number of bacteria that are resistant to this antibiotic"
    DrugMtbStat _mdr;                               // "Number of bacteria that are resistant to INH and RIF"
    DrugMtbStat _compensatedResistant;              // "Number of resistant bacteria that have compensatory mutations for this antibiotic"
    DrugMtbStat _mdrCompensatedResistant;           // "Number of MDR bacteria that have compensatory mutations for this antibiotic"
    DrugMtbStat _mdrDoubleCompensatedResistant;     // "Number of MDR bacteria that have compensatory mutations for both antibiotics"
    DrugMtbStat _cumResistant;                      // "Cumulative number of bacteria that are resistant to this antibiotic"
    DrugMtbStat _cumMdr;                            // "Cumulative number of bacteria that are resistant to INH and RIF"
    DrugMtbStat _cumCompensated;                    // "Cumulative number of resistant bacteria that have compensatory mutations for this antibiotic"
    DrugMtbStat _cumMdrCompensatedResistant;        // "Cumulative number of MDR bacteria that have compensatory mutations for this antibiotic"
    DrugMtbStat _cumMdrDoubleCompensated;           // "Cumulative number of MDR bacteria that have compensatory mutations for both antibiotics"
    DrugMtbStat _effluxInduced;                     // "Number of bacteria with efflux pumps induced"
    DrugMtbStat _doubleEffluxInduced;               // "Number of bacteria with efflux pumps induced for both antibiotics"

    // Total drug concentration on the grid is kept for a number of time steps
    // specified as a parameter file parameter, so a Boost array cannot be used
    // because that requires specifying the size at compile time.
    std::vector<Scalar> _drugConcentration;          // mg/L
    std::vector<Scalar> _drugConcentrationNorm;      // mg/L
    std::vector<Scalar> _drugConcentrationCaseum;      // mg/L
    std::vector<Scalar> _drugConcentrationGran;      // mg/L
    std::vector<Scalar> _drugConcentrationblood;     // mol/L
    std::vector<Scalar> _drugConcentrationLung;      // mg/L
    std::vector<Scalar> _drugConcentrationPeriph;    // mg/kg

    GrAccumulator<double> _drugDistStat;
protected:
    AntibioticSimulationGrid();
    void initFromParameterFile(const boost::property_tree::ptree& pt);

public:
    AntibioticSimulationGrid(const string& antibioticName,
                             const Pos& dim,
                             const boost::property_tree::ptree& pt,
                             GrDiffusion* pDiffusion, double dt);
    AntibioticSimulationGrid(const AntibioticSimulationGrid& asg);
    virtual ~AntibioticSimulationGrid();

    AntibioticSimulationGrid* clone() const
    {
        return new AntibioticSimulationGrid(*this);
    }

    void loadParameters(const boost::property_tree::ptree& pt, double dt);
    void setupDiffusion(GrDiffusion* pDiffusion, Pos dim);
    template <typename T>
    bool getDrugParam(const boost::property_tree::ptree& drugParams,
                        const string& antibioticName, const string& paramName,
                        T& paramValue);
    double getDrugParamFromLogNormalDistribution(const boost::property_tree::ptree& drugParams,
                        const string& antibioticName, const string& paramName);
    // Resistance stats
    void updateResistanceStats(MtbGroupLocation location);
    void updateMdrStats(MtbGroupLocation location);
    void updateCompensatoryMutationStats(MtbGroupLocation location);
    void updateMdrCompensatoryMutationStats(MtbGroupLocation location);
    void updateMdrDoubleCompensatoryMutationStats(MtbGroupLocation location);
    // Cumulative resistance stats
    void updateCumulativeResistanceStats(MtbGroupLocation location);
    void updateCumulativeMdrStats(MtbGroupLocation location);
    void updateCumulativeCompensatoryMutationStats(MtbGroupLocation location);
    void updateCumulativeMdrCompensatoryMutationStats(MtbGroupLocation location);
    void updateCumulativeMdrDoubleCompensatoryMutationStats(MtbGroupLocation location);
    // Efflux pump stats
    void updateDoubleEffluxInducedStats(MtbGroupLocation location);
    void updateEffluxInducedStats(MtbGroupLocation location);


    void partitionAntibiotic(GrGrid &pg, unsigned abxInd,
                             int numMolecularPerDiffusion,
                             double timestepMolecular,
                             ODESolvers::ODEMethod method);
    void partitionAntibioticSteadyState(GrGrid& pg, unsigned abxInd);
    void partitionAntibioticDynamic(GrGrid& pg, unsigned abxInd,
                                    int numMolecularPerDiffusion,
                                    double timestepMolecular,
                                    ODESolvers::ODEMethod method);
    void updateAntibioticPartitioningODEs(AntibioticConcentration* macAbxConc, Pos p, bool hasMac,
                                          double caseumLevel, double kfc, double krc, double kfe, double kre,
                                          double kin, double kout, double dt, ODESolvers::ODEMethod odeSolver) ;
    void writeValarrayFromMembersCaseumBinding(valarray<double>& inputVector, double freeAbx, double casBndAbx, double epithBndAbx, double intraAbx);
    void writeMembersFromValarrayCaseumBinding(valarray<double>& inputVector, AntibioticConcentration *macAbxConc, Pos p, bool hasMac);
    void printMembers(AntibioticConcentration* macAbxConc, Pos p, bool hasMac);

    // Member variables
    string getName() const;
    int getSize() const;

    const double& pmat(const Pos& p) const;
    void setPmat(const Pos& p, double v);
    void incPmat(const Pos& p, double v);
    const vector<double>& pmat() const;
    vector<double> &pmat();

    void swapGrids();
    void updateNonDiffusingGrids();

    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    CrankNicolsonADIChemical& getCnadiChemical();
    #endif

    AntibioticGrid& currentGrid();
    AntibioticGrid& nextGrid();

    void setCurrentGrid(AntibioticGrid* antibioticGrid);
    void setNextGrid(AntibioticGrid* antibioticGrid);

	void setEntireGrid(double value);

    // Vascular variables get/set functions
    double& getBloodConcentration();
    double& getLungConcentration();
    double& getPeriphConcentration();
    double &getIntMacConcentration();
    void setBloodConcentration(double bc);
    void incBloodConcentration(double bc);
    void setabsComp1Concentration(double ac);
    double& getabsComp1Concentration();
    void incabsComp1Concentration(double ac);
    void setabsComp2Concentration(double ac);
    double& getabsComp2Concentration();
    void incabsComp2Concentration(double ac);
    void setInitVector(std::valarray<double> concvec);
    const std::valarray<double>& getInitVector() const;
    std::valarray<double>& getInitVector();

    // Stats Grids get/set functions.
    const double& intracellularGrid(const Pos& p) const;
    const double* intracellularGrid() const;
    double& intracellularGrid(const Pos& p);
    double* intracellularGrid();
    void setIntracellularGrid(const Pos& p, double v);

    const double& totalGrid(const Pos& p) const;
    const double* totalGrid() const;
    double& totalGrid(const Pos& p);
    double* totalGrid();
    void setTotalGrid(const Pos& p, double v);
    const std::vector<double>& totalGridVector() const;
    std::vector<double>& totalGridVector();

    const double& aucIntraGrid(const Pos& p) const;
    const double* aucIntraGrid() const;
    double& aucIntraGrid(const Pos& p);
    double* aucIntraGrid();
    void setAucIntraGrid(const Pos& p, double v);
    void incAucIntraGrid(const Pos& p, double v);

    const double& aucExtraGrid(const Pos& p) const;
    const double* aucExtraGrid() const;
    double& aucExtraGrid(const Pos& p);
    double* aucExtraGrid();
    void setAucExtraGrid(const Pos& p, double v);
    void incAucExtraGrid(const Pos& p, double v);

    const double& aucTotalGrid(const Pos& p) const;
    const double* aucTotalGrid() const;
    double& aucTotalGrid(const Pos& p);
    double* aucTotalGrid();
    void setAucTotalGrid(const Pos& p, double v);
    void incAucTotalGrid(const Pos& p, double v);

    const double& cmaxIntraGrid(const Pos& p) const;
    const double* cmaxIntraGrid() const;
    double& cmaxIntraGrid(const Pos& p);
    double* cmaxIntraGrid();
    void setCmaxIntraGrid(const Pos& p, double v);

    const double& cmaxExtraGrid(const Pos& p) const;
    const double* cmaxExtraGrid() const;
    double& cmaxExtraGrid(const Pos& p);
    double* cmaxExtraGrid();
    void setCmaxExtraGrid(const Pos& p, double v);

    const double& cmaxTotalGrid(const Pos& p) const;
    const double* cmaxTotalGrid() const;
    double& cmaxTotalGrid(const Pos& p);
    double* cmaxTotalGrid();
    void setCmaxTotalGrid(const Pos& p, double v);

    //Stats get/set functions
    void resetStats();

    const double& getTotalWholeGrid() const;
    double& getTotalWholeGrid() ;
    void setTotalWholeGrid(const double& v) ;

    const double& getTotalGran() const;
    double& getTotalGran() ;
    void setTotalGran(const double& v) ;

    const double& getTotalNorm() const;
    double& getTotalNorm() ;
    void setTotalNorm(const double& v);

    const double& getTotalCaseum() const;
    double& getTotalCaseum() ;
    void setTotalCaseum(const double& v);

    const double& getPartialCaseum() const;
    double& getPartialCaseum() ;
    void setPartialCaseum(const double& v);

    // AvgConcInfInt gives the average antibiotic concentration in
    // infected macrophages
    const double& getAvgConcInfInt() const;
    double& getAvgConcInfInt() ;
    void setAvgConcInfInt(const double& v) ;

    // AvgConcExtMtb gives the average antibiotic concentration in grid
    // compartments with replicating extracellular Mtb
    const double& getAvgConcExtMtb() const;
    double& getAvgConcExtMtb() ;
    void setAvgConcExtMtb(const double& v) ;

    // AvgConcNonMtb gives the avearge antibiotic concentration in grid
    // compartments with non-replicating extracellular Mtb
    const double& getAvgConcNonMtb() const;
    double& getAvgConcNonMtb() ;
    void setAvgConcNonMtb(const double& v) ;

    const double& getDrugConcentrationArea() const;
    double& getDrugConcentrationArea() ;
    void setDrugConcentrationArea(const double& v) ;

    const double& getDrugConcentrationAreaNorm() const;
    double& getDrugConcentrationAreaNorm() ;
    void setDrugConcentrationAreaNorm(const double& v) ;

    const double& getDrugConcentrationAreaCaseum() const;
    double& getDrugConcentrationAreaCaseum() ;
    void setDrugConcentrationAreaCaseum(const double& v) ;

    const double& getDrugConcentrationAreaGran() const;
    double& getDrugConcentrationAreaGran() ;
    void setDrugConcentrationAreaGran(const double& v);

    const double& getDrugConcentrationAreaBlood() const;
    double& getDrugConcentrationAreaBlood() ;
    void setDrugConcentrationAreaBlood(const double& v) ;

    const double& getDrugConcentrationAreaLung() const;
    double& getDrugConcentrationAreaLung() ;
    void setDrugConcentrationAreaLung(const double& v);

    const double& getDrugConcentrationAreaPeriph() const;
    double& getDrugConcentrationAreaPeriph();
    void setDrugConcentrationAreaPeriph(const double& v) ;

    const double& getCmax() const;
    double& getCmax() ;
    void setCmax(const double& v) ;

    const double& getCmaxnorm() const;
    double& getCmaxnorm() ;
    void setCmaxnorm(const double& v) ;

    const double& getCmaxcaseum() const;
    double& getCmaxcaseum() ;
    void setCmaxcaseum(const double& v) ;

    const double& getCmaxgran() const;
    double& getCmaxgran() ;
    void setCmaxgran(const double& v) ;

    const double& getCmaxblood() const;
    double& getCmaxblood();
    void setCmaxblood(const double& v) ;

    const double& getCompOverMIC() const;
    double& getCompOverMIC();
    void setCompOverMIC(const double& v) ;

    const double& getExtMtbOverMIC() const;
    double& getExtMtbOverMIC() ;
    void setExtMtbOverMIC(const double& v) ;

    const double& getMacsOverMIC() const;
    double& getMacsOverMIC() ;
    void setMacsOverMIC(const double& v);

    const double& getIntMtbOverMIC() const;
    double& getIntMtbOverMIC() ;
    void setIntMtbOverMIC(const double& v) ;

    const double& getIntMtbHighRiskOfResistance() const;
    double& getIntMtbHighRiskOfResistance() ;
    void setIntMtbHighRiskOfResistance(const double& v) ;

    const double& getExtMtbHighRiskOfResistance() const;
    double& getExtMtbHighRiskOfResistance() ;
    void setExtMtbHighRiskOfResistance(const double& v) ;

    const double& getNonrepMtbHighRiskOfResistance() const;
    double& getNonrepMtbHighRiskOfResistance();
    void setNonrepMtbHighRiskOfResistance(const double& v) ;

    const double& getMacsHighRiskOfResistance() const;
    double& getMacsHighRiskOfResistance() ;
    void setMacsHighRiskOfResistance(const double& v) ;

    const double& getCompHighRiskOfResistance() const;
    double& getCompHighRiskOfResistance() ;
    void setCompHighRiskOfResistance(const double& v) ;

    const double& getIntMtbSingleDrugKill() const;
    double& getIntMtbSingleDrugKill() ;
    void setIntMtbSingleDrugKill(const double& v);
    void incIntMtbSingleDrugKill(const double& v);

    const double& getExtMtbSingleDrugKill() const;
    double& getExtMtbSingleDrugKill() ;
    void setExtMtbSingleDrugKill(const double& v);
    void incExtMtbSingleDrugKill(const double& v);

    const double& getNonRepSingleDrugKill() const;
    double& getNonRepSingleDrugKill() ;
    void setNonRepSingleDrugKill(const double& v);
    void incNonRepSingleDrugKill(const double& v);

    // Resistance stats
    DrugMtbStat& resistant();
    DrugMtbStat& mdr();
    DrugMtbStat& compensatedResistant();
    DrugMtbStat& mdrCompensatedResistant();
    DrugMtbStat& mdrDoubleCompensatedResistant();
    DrugMtbStat& cumResistant();
    DrugMtbStat& cumMdr();
    DrugMtbStat& cumCompensated();
    DrugMtbStat& cumMdrCompensatedResistant();
    DrugMtbStat& cumMdrDoubleCompensated();
    DrugMtbStat& effluxInduced();
    DrugMtbStat& doubleEffluxInduced();

    // Cumulative concentration stats
    Scalar& getDrugConcentration(size_t i);
    Scalar getDrugConcentration(size_t i) const;
    const std::vector<Scalar>& getDrugConcentration() const;
    std::vector<Scalar>& getDrugConcentration() ;
    void setDrugConcentration(size_t i, const Scalar v);

    Scalar& getDrugConcentrationNorm(size_t i);
    Scalar getDrugConcentrationNorm(size_t i) const;
    const std::vector<Scalar>& getDrugConcentrationNorm() const;
    std::vector<Scalar>& getDrugConcentrationNorm();
    void setDrugConcentrationNorm(size_t i, const Scalar v);

    Scalar& getDrugConcentrationCaseum(size_t i);
    Scalar getDrugConcentrationCaseum(size_t i) const;
    const std::vector<Scalar>& getDrugConcentrationCaseum() const;
    std::vector<Scalar>& getDrugConcentrationCaseum();
    void setDrugConcentrationCaseum(size_t i, const Scalar v);

    Scalar& getDrugConcentrationGran(size_t i);
    Scalar getDrugConcentrationGran(size_t i) const;
    const std::vector<Scalar>& getDrugConcentrationGran() const;
    std::vector<Scalar>& getDrugConcentrationGran();
    void setDrugConcentrationGran(size_t i, const Scalar v);

    Scalar& getDrugConcentrationBlood(size_t i);
    Scalar getDrugConcentrationBlood(size_t i) const;
    const std::vector<Scalar>& getDrugConcentrationBlood() const;
    std::vector<Scalar>& getDrugConcentrationBlood() ;
    void setDrugConcentrationBlood(size_t i, const Scalar v);

    Scalar& getDrugConcentrationLung(size_t i);
    Scalar getDrugConcentrationLung(size_t i) const;
    const std::vector<Scalar>& getDrugConcentrationLung() const;
    std::vector<Scalar>& getDrugConcentrationLung() ;
    void setDrugConcentrationLung(size_t i, const Scalar v);

    Scalar& getDrugConcentrationPeriph(size_t i);
    Scalar getDrugConcentrationPeriph(size_t i) const;
    const std::vector<Scalar>& getDrugConcentrationPeriph() const;
    std::vector<Scalar>& getDrugConcentrationPeriph();
    void setDrugConcentrationPeriph(size_t i, const Scalar v);

    GrAccumulator<double>& getdrugDistStat() ;
    const GrAccumulator<double>& getdrugDistStat() const ;

    // Parameter get functions
    const int& PKModel() const;
    int& PKModel();

    const double& vesselActivity() const;
    double& vesselActivity() ;

    const int& partitionUsingSteadyState() const;
    int &partitionUsingSteadyState();

    const double& degRateConst() const;
    double& degRateConst() ;

    const double& degRateConstInt() const;
    double& degRateConstInt() ;

    const double& diffusivity() const;
    double& diffusivity() ;

    const double& minDiffusivity() const;
    double& minDiffusivity() ;

    const double& kAbs() const;
    double& kAbs() ;

    const double& q() const;
    double& q() ;

    const double& volDistCentral() const;
    double& volDistCentral() ;

    const double& volDistPeriph() const;
    double& volDistPeriph() ;

    const double& clearance() const;
    double& clearance() ;

    const double& dose() const;
    double& dose() ;

    const int &dosageInterval() const;
    int& dosageInterval() ;

    const double& cellUptake() const;
    double& cellUptake() ;

    const double& vascularPermeability() const;
    double& vascularPermeability() ;

    const double& permCoeff() const;
    double& permCoeff() ;

    const double& intraActivity() const;
    double& intraActivity() ;

    const double& extraActivity() const;
    double& extraActivity() ;

    const double& nonrepActivity() const;
    double& nonrepActivity() ;

    const double& c50Extra() const;
    double& c50Extra() ;

    const double& c50Intra() const;
    double& c50Intra() ;

    const double& c50Nonrep() const;
    double& c50Nonrep() ;

    const double& consumptionRate() const;
    double& consumptionRate() ;

    const double& consumptionHalfSat() const;
    double& consumptionHalfSat() ;

    const double& mic() const;
    double& mic() ;

    const double& hillExtra() const;
    double& hillExtra() ;

    const double& hillIntra() const;
    double& hillIntra() ;

    const double& hillNonrep() const;
    double& hillNonrep() ;

    const double& caseumUnboundFraction() const;
    double& caseumUnboundFraction() ;

    const double& caseumBindingRateConst() const;
    double& caseumBindingRateConst() ;

    const double& epitheliumBindingKa() const;
    double& epitheliumBindingKa() ;

    const double& epitheliumBindingRateConst() const;
    double& epitheliumBindingRateConst() ;

    const double& cellularKout() const;
    double& cellularKout() ;

    const double& MW() const;
    double& MW() ;

    const int &resistRiskIndicatorFlag() const;
    int& resistRiskIndicatorFlag() ;

    const double &numberResistMutations() const;
    double& numberResistMutations();

    const double& relativeFitnessResist() const;
    double &relativeFitnessResist();

    const double &numberCompensatoryMutations() const;
    double& numberCompensatoryMutations();

    const double& foldIncreaseC50FromResistance() const;
    double& foldIncreaseC50FromResistance();

    const int& numTimeStepsEffluxInduction() const;
    int& numTimeStepsEffluxInduction();

    const double& minEffluxInduction() const;
    double& minEffluxInduction();

    const double& maxEffluxInduction() const;
    double& maxEffluxInduction();

    const double& concentrationChangeByEfflux() const;
    double& concentrationChangeByEfflux();

    const int& numberOfRandomMonoResistantBugsToAdd() const;
    int& numberOfRandomMonoResistantBugsToAdd();

    const double& volDistLung() const;
    double& volDistLung();

    double calculateMicFromC50(double c50, double hill, double fractionKill);

    //Partitioning ODES
    static std::auto_ptr<ODESolvers::Stepper> stepper; // ODE Stepper for all agents (unless overridden)
    static std::auto_ptr<GridAntibioticPartitioningFunc> deriv; // ODE Function for all agents (unless overridden)
    virtual ODESolvers::Stepper* getStepper(ODESolvers::ODEMethod odeSolver, int size)
    {
      static std::auto_ptr<ODESolvers::Stepper> stepper;
      static ODESolvers::ODEMethod initMethod = odeSolver;
      if(stepper.get() == NULL || odeSolver != initMethod)
        {
          stepper.reset(ODESolvers::StepperFactory(odeSolver, size));
          initMethod = odeSolver;
        }
      return stepper.get();
    }
    static GridAntibioticPartitioningFunc* buildDerivFunc();
    virtual GridAntibioticPartitioningFunc* getDerivFunc()
    {
        static std::auto_ptr<GridAntibioticPartitioningFunc> deriv; // ODE Function for all agents (unless overridden)
        if(deriv.get() == NULL)
          {
            deriv.reset(buildDerivFunc());
          }
        return deriv.get();
    }

    // Risk of drug resistance
    void getResistRiskValues(int riskFlag, double& riskIndicatorValueIntra, double& riskIndicatorValueExtra,
                             double& riskLowerBound, double& riskUpperBound, const Pos &p);

    double getKillThreshold(double percThreshold) const;

private:
    friend class boost::serialization::access;  

    template<class Archive>
    void serialize(Archive& ar, const unsigned int version);
};


// Member variables
inline string AntibioticSimulationGrid::getName() const
{
   return _name;
}

inline int AntibioticSimulationGrid::getSize() const
{
  return _dim.size();
}

inline const double& AntibioticSimulationGrid::pmat(const Pos& p) const
{
    return _pMat[GridSpec::padInd(_dim, p)];
}

inline void AntibioticSimulationGrid::setPmat(const Pos& p, double v)
{
    _pMat[GridSpec::padInd(_dim, p)] = v;
}
inline void AntibioticSimulationGrid::incPmat(const Pos& p, double v)
{
    _pMat[GridSpec::padInd(_dim, p)] += v;
}
inline const vector<double> &AntibioticSimulationGrid::pmat() const
{
    return _pMat;
}
inline vector<double>& AntibioticSimulationGrid::pmat()
{
    return _pMat;
}
// Swap the current and next grids.
// Done after the next state has been stored in the next grid based
// on the current state in the current.
inline void AntibioticSimulationGrid::swapGrids()
{
    std::swap(_currentGrid, _nextGrid);
}

// In order for non-diffusing grids not to get lost in a swap, write
// current grid values for those grids to next grids
inline void AntibioticSimulationGrid::updateNonDiffusingGrids()
{
    GridSpec::EntireGridIterator egi = GridSpec::egiBegin(_dim);
    GridSpec::EntireGridIterator egiEnd =  GridSpec::egiEnd(_dim);

    for(; egi != egiEnd; ++egi)
    {
        const Pos& pos = *egi;
        // Caseum bound grid
        double currentValue = currentGrid().casBnd_grid(pos);
        nextGrid().setCasBnd_grid(pos, currentValue);
        // Epith bound grid
        currentValue = currentGrid().epithBnd_grid(pos);
        nextGrid().setEpithBnd_grid(pos, currentValue);

    }

}

#if DIMENSION == 2
// Crank-Nicolson ADI is a 2D only diffusion algorithm.
inline CrankNicolsonADIChemical& AntibioticSimulationGrid::getCnadiChemical()
{
    return *_cnadiChemical;
}
#endif

inline AntibioticGrid& AntibioticSimulationGrid::currentGrid()
{
    return *_currentGrid;
}

inline AntibioticGrid& AntibioticSimulationGrid::nextGrid()
{
    return *_nextGrid;
}

inline void AntibioticSimulationGrid::setCurrentGrid(AntibioticGrid* antibioticGrid)
{
  _currentGrid = antibioticGrid;
}

inline void AntibioticSimulationGrid::setNextGrid(AntibioticGrid* antibioticGrid)
{
  _nextGrid = antibioticGrid;
}

inline void AntibioticSimulationGrid::setEntireGrid(double value)
{
    _currentGrid->setEntireGrid(value);
}

// Vascular variables get/set functions
inline double &AntibioticSimulationGrid::getBloodConcentration()
{
    return _bloodConcentration;
}
inline double &AntibioticSimulationGrid::getLungConcentration()
{
    return _lungCompConcentration;  // mg/kg
}
inline void AntibioticSimulationGrid::setInitVector(std::valarray<double> concvec)
{
    _initvector = concvec;
}
inline const std::valarray<double>& AntibioticSimulationGrid::getInitVector() const
{
    return _initvector;
}
inline std::valarray<double>& AntibioticSimulationGrid::getInitVector()
{
    return _initvector;
}
inline void AntibioticSimulationGrid::setBloodConcentration(double bc)
{
    _bloodConcentration = bc;
}
inline void AntibioticSimulationGrid::incBloodConcentration(double bc)
{
    _bloodConcentration += bc;
}
inline void AntibioticSimulationGrid::setabsComp1Concentration(double ac)
{
    _absComp1Concentration = ac;
}
inline double &AntibioticSimulationGrid::getabsComp1Concentration()
{
    return _absComp1Concentration;
}
inline void AntibioticSimulationGrid::incabsComp1Concentration(double ac)
{
    _absComp1Concentration += ac;
}
inline void AntibioticSimulationGrid::setabsComp2Concentration(double ac)
{
    _absComp2Concentration = ac;
}
inline double &AntibioticSimulationGrid::getabsComp2Concentration()
{
    return _absComp2Concentration;
}
inline void AntibioticSimulationGrid::incabsComp2Concentration(double ac)
{
    _absComp2Concentration += ac;
}

inline double &AntibioticSimulationGrid::getPeriphConcentration()
{
    return _perCompConcentration;
}
inline  double &AntibioticSimulationGrid::getIntMacConcentration()
{
    return _intMacConcentration;
}
// END Vascular variables get/set functions

//Stats Grids get/set functions
inline const double& AntibioticSimulationGrid::intracellularGrid(const Pos& p) const
{
    return _intracellularGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::intracellularGrid() const
{
    return _intracellularGrid.data();
}
inline double& AntibioticSimulationGrid::intracellularGrid(const Pos& p)
{
    return _intracellularGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::intracellularGrid()
{
    return _intracellularGrid.data();
}
inline void AntibioticSimulationGrid::setIntracellularGrid(const Pos& p, double v)
{
    _intracellularGrid[GridSpec::ind(_dim, p)] = v;
}


inline const double& AntibioticSimulationGrid::totalGrid(const Pos& p) const
{
    return _totalGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::totalGrid() const
{
    return _totalGrid.data();
}
inline double& AntibioticSimulationGrid::totalGrid(const Pos& p)
{
    return _totalGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::totalGrid()
{
    return _totalGrid.data();
}
inline void AntibioticSimulationGrid::setTotalGrid(const Pos& p, double v)
{
    _totalGrid[GridSpec::ind(_dim, p)] = v;
}
inline const std::vector<double>& AntibioticSimulationGrid::totalGridVector() const
{
    return _totalGrid;
}

inline std::vector<double>& AntibioticSimulationGrid::totalGridVector()
{
    return _totalGrid;
}

inline const double& AntibioticSimulationGrid::aucIntraGrid(const Pos& p) const
{
    return _aucIntraGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::aucIntraGrid() const
{
    return _aucIntraGrid.data();
}
inline double& AntibioticSimulationGrid::aucIntraGrid(const Pos& p)
{
    return _aucIntraGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::aucIntraGrid()
{
    return _aucIntraGrid.data();
}
inline void AntibioticSimulationGrid::setAucIntraGrid(const Pos& p, double v)
{
    _aucIntraGrid[GridSpec::ind(_dim, p)] = v;
}
inline void AntibioticSimulationGrid::incAucIntraGrid(const Pos& p, double v)
{
    _aucIntraGrid[GridSpec::ind(_dim, p)] += v;
}

inline const double& AntibioticSimulationGrid::aucExtraGrid(const Pos& p) const
{
    return _aucExtraGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::aucExtraGrid() const
{
    return _aucExtraGrid.data();
}
inline double& AntibioticSimulationGrid::aucExtraGrid(const Pos& p)
{
    return _aucExtraGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::aucExtraGrid()
{
    return _aucExtraGrid.data();
}
inline void AntibioticSimulationGrid::setAucExtraGrid(const Pos& p, double v)
{
    _aucExtraGrid[GridSpec::ind(_dim, p)] = v;
}
inline void AntibioticSimulationGrid::incAucExtraGrid(const Pos& p, double v)
{
    _aucExtraGrid[GridSpec::ind(_dim, p)] += v;
}

inline const double& AntibioticSimulationGrid::aucTotalGrid(const Pos& p) const
{
    return _aucTotalGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::aucTotalGrid() const
{
    return _aucTotalGrid.data();
}
inline double& AntibioticSimulationGrid::aucTotalGrid(const Pos& p)
{
    return _aucTotalGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::aucTotalGrid()
{
    return _aucTotalGrid.data();
}
inline void AntibioticSimulationGrid::setAucTotalGrid(const Pos& p, double v)
{
    _aucTotalGrid[GridSpec::ind(_dim, p)] = v;
}
inline void AntibioticSimulationGrid::incAucTotalGrid(const Pos& p, double v)
{
    _aucTotalGrid[GridSpec::ind(_dim, p)] += v;
}


inline const double& AntibioticSimulationGrid::cmaxIntraGrid(const Pos& p) const
{
    return _cmaxIntraGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::cmaxIntraGrid() const
{
    return _cmaxIntraGrid.data();
}
inline double& AntibioticSimulationGrid::cmaxIntraGrid(const Pos& p)
{
    return _cmaxIntraGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::cmaxIntraGrid()
{
    return _cmaxIntraGrid.data();
}
inline void AntibioticSimulationGrid::setCmaxIntraGrid(const Pos& p, double v)
{
    _cmaxIntraGrid[GridSpec::ind(_dim, p)] = v;
}

inline const double& AntibioticSimulationGrid::cmaxExtraGrid(const Pos& p) const
{
    return _cmaxExtraGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::cmaxExtraGrid() const
{
    return _cmaxExtraGrid.data();
}
inline double& AntibioticSimulationGrid::cmaxExtraGrid(const Pos& p)
{
    return _cmaxExtraGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::cmaxExtraGrid()
{
    return _cmaxExtraGrid.data();
}
inline void AntibioticSimulationGrid::setCmaxExtraGrid(const Pos& p, double v)
{
    _cmaxExtraGrid[GridSpec::ind(_dim, p)] = v;
}


inline const double& AntibioticSimulationGrid::cmaxTotalGrid(const Pos& p) const
{
    return _cmaxTotalGrid[GridSpec::ind(_dim, p)];
}
inline const double* AntibioticSimulationGrid::cmaxTotalGrid() const
{
    return _cmaxTotalGrid.data();
}
inline double& AntibioticSimulationGrid::cmaxTotalGrid(const Pos& p)
{
    return _cmaxTotalGrid[GridSpec::ind(_dim, p)];
}
inline double* AntibioticSimulationGrid::cmaxTotalGrid()
{
    return _cmaxTotalGrid.data();
}
inline void AntibioticSimulationGrid::setCmaxTotalGrid(const Pos& p, double v)
{
    _cmaxTotalGrid[GridSpec::ind(_dim, p)] = v;
}
// END Stats Grids get/set functions

//Stats get set functions
inline void AntibioticSimulationGrid::resetStats()
{
    // Reset all resetable stats
    _totalWholeGrid = 0.0;
    _totalGran = 0.0;
    _totalNorm = 0.0;
    _totalCaseum = 0.0;
    _partialCaseum = 0.0;
    _avgConcInfInt = 0.0;
    _avgConcExtMtb = 0.0;
    _avgConcNonMtb = 0.0;

    _compOverMIC = 0.0;
    _extMtbOverMIC = 0.0;
    _macsOverMIC = 0.0;
    _intMtbOverMIC = 0.0;
    _intMtbHighRiskOfResistance = 0.0;
    _extMtbHighRiskOfResistance = 0.0;
    _nonrepMtbHighRiskOfResistance = 0.0;
    _macsHighRiskOfResistance = 0.0;
    _compHighRiskOfResistance = 0.0;

    _resistant.reset();
    _mdr.reset();
    _compensatedResistant.reset();
    _mdrCompensatedResistant.reset();
    _mdrDoubleCompensatedResistant.reset();
    _effluxInduced.reset();
    _doubleEffluxInduced.reset();

    _drugDistStat.reset();

}

inline const double& AntibioticSimulationGrid::getTotalWholeGrid() const { return _totalWholeGrid; }
inline double& AntibioticSimulationGrid::getTotalWholeGrid() { return _totalWholeGrid; }
inline void AntibioticSimulationGrid::setTotalWholeGrid(const double& v) {_totalWholeGrid = v;}

inline const double& AntibioticSimulationGrid::getTotalGran() const { return _totalGran; }
inline double& AntibioticSimulationGrid::getTotalGran() { return _totalGran; }
inline void AntibioticSimulationGrid::setTotalGran(const double& v) {_totalGran = v;}

inline const double& AntibioticSimulationGrid::getTotalNorm() const { return _totalNorm; }
inline double& AntibioticSimulationGrid::getTotalNorm() { return _totalNorm; }
inline void AntibioticSimulationGrid::setTotalNorm(const double& v) {_totalNorm = v;}

inline const double& AntibioticSimulationGrid::getTotalCaseum() const { return _totalCaseum; }
inline double& AntibioticSimulationGrid::getTotalCaseum() { return _totalCaseum; }
inline void AntibioticSimulationGrid::setTotalCaseum(const double& v) {_totalCaseum = v;}

inline const double& AntibioticSimulationGrid::getPartialCaseum() const { return _partialCaseum; }
inline double& AntibioticSimulationGrid::getPartialCaseum() { return _partialCaseum; }
inline void AntibioticSimulationGrid::setPartialCaseum(const double& v) {_partialCaseum = v;}

inline const double& AntibioticSimulationGrid::getAvgConcInfInt() const { return _avgConcInfInt; }
inline double& AntibioticSimulationGrid::getAvgConcInfInt() { return _avgConcInfInt; }
inline void AntibioticSimulationGrid::setAvgConcInfInt(const double& v) {_avgConcInfInt = v;}

inline const double& AntibioticSimulationGrid::getAvgConcExtMtb() const { return _avgConcExtMtb; }
inline double& AntibioticSimulationGrid::getAvgConcExtMtb() { return _avgConcExtMtb; }
inline void AntibioticSimulationGrid::setAvgConcExtMtb(const double& v) {_avgConcExtMtb = v;}

inline const double& AntibioticSimulationGrid::getAvgConcNonMtb() const { return _avgConcNonMtb; }
inline double& AntibioticSimulationGrid::getAvgConcNonMtb() { return _avgConcNonMtb; }
inline void AntibioticSimulationGrid::setAvgConcNonMtb(const double& v) {_avgConcNonMtb = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationArea() const { return _drugConcentrationArea; }
inline double& AntibioticSimulationGrid::getDrugConcentrationArea() { return _drugConcentrationArea; }
inline void AntibioticSimulationGrid::setDrugConcentrationArea(const double& v) {_drugConcentrationArea = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationAreaNorm() const { return _drugConcentrationAreaNorm; }
inline double& AntibioticSimulationGrid::getDrugConcentrationAreaNorm() { return _drugConcentrationAreaNorm; }
inline void AntibioticSimulationGrid::setDrugConcentrationAreaNorm(const double& v) {_drugConcentrationAreaNorm = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationAreaCaseum() const { return _drugConcentrationAreaCaseum; }
inline double& AntibioticSimulationGrid::getDrugConcentrationAreaCaseum() { return _drugConcentrationAreaCaseum; }
inline void AntibioticSimulationGrid::setDrugConcentrationAreaCaseum(const double& v) {_drugConcentrationAreaCaseum = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationAreaGran() const { return _drugConcentrationAreaGran; }
inline double& AntibioticSimulationGrid::getDrugConcentrationAreaGran() { return _drugConcentrationAreaGran; }
inline void AntibioticSimulationGrid::setDrugConcentrationAreaGran(const double& v) {_drugConcentrationAreaGran = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationAreaBlood() const { return _drugConcentrationAreaBlood; }
inline double& AntibioticSimulationGrid::getDrugConcentrationAreaBlood() { return _drugConcentrationAreaBlood; }
inline void AntibioticSimulationGrid::setDrugConcentrationAreaBlood(const double& v) {_drugConcentrationAreaBlood = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationAreaLung() const { return _drugConcentrationAreaLung; }
inline double& AntibioticSimulationGrid::getDrugConcentrationAreaLung() { return _drugConcentrationAreaLung; }
inline void AntibioticSimulationGrid::setDrugConcentrationAreaLung(const double& v) {_drugConcentrationAreaLung = v;}

inline const double& AntibioticSimulationGrid::getDrugConcentrationAreaPeriph() const { return _drugConcentrationAreaPeriph; }
inline double& AntibioticSimulationGrid::getDrugConcentrationAreaPeriph() { return _drugConcentrationAreaPeriph; }
inline void AntibioticSimulationGrid::setDrugConcentrationAreaPeriph(const double& v) {_drugConcentrationAreaPeriph = v;}

inline const double& AntibioticSimulationGrid::getCmax() const { return _cmax; }
inline double& AntibioticSimulationGrid::getCmax() { return _cmax; }
inline void AntibioticSimulationGrid::setCmax(const double& v) {_cmax = v;}

inline const double& AntibioticSimulationGrid::getCmaxnorm() const { return _cmaxNorm; }
inline double& AntibioticSimulationGrid::getCmaxnorm() { return _cmaxNorm; }
inline void AntibioticSimulationGrid::setCmaxnorm(const double& v) {_cmaxNorm = v;}

inline const double& AntibioticSimulationGrid::getCmaxcaseum() const { return _cmaxCaseum; }
inline double& AntibioticSimulationGrid::getCmaxcaseum() { return _cmaxCaseum; }
inline void AntibioticSimulationGrid::setCmaxcaseum(const double& v) {_cmaxCaseum = v;}

inline const double& AntibioticSimulationGrid::getCmaxgran() const { return _cmaxGran; }
inline double& AntibioticSimulationGrid::getCmaxgran() { return _cmaxGran; }
inline void AntibioticSimulationGrid::setCmaxgran(const double& v) {_cmaxGran = v;}

inline const double& AntibioticSimulationGrid::getCmaxblood() const { return _cmaxBlood; }
inline double& AntibioticSimulationGrid::getCmaxblood() { return _cmaxBlood; }
inline void AntibioticSimulationGrid::setCmaxblood(const double& v) {_cmaxBlood = v;}

inline const double& AntibioticSimulationGrid::getCompOverMIC() const { return _compOverMIC; }
inline double& AntibioticSimulationGrid::getCompOverMIC() { return _compOverMIC; }
inline void AntibioticSimulationGrid::setCompOverMIC(const double& v) {_compOverMIC = v;}

inline const double& AntibioticSimulationGrid::getExtMtbOverMIC() const { return _extMtbOverMIC; }
inline double& AntibioticSimulationGrid::getExtMtbOverMIC() { return _extMtbOverMIC; }
inline void AntibioticSimulationGrid::setExtMtbOverMIC(const double& v) {_extMtbOverMIC = v;}

inline const double& AntibioticSimulationGrid::getMacsOverMIC() const { return _macsOverMIC; }
inline double& AntibioticSimulationGrid::getMacsOverMIC() { return _macsOverMIC; }
inline void AntibioticSimulationGrid::setMacsOverMIC(const double& v) {_macsOverMIC = v;}

inline const double& AntibioticSimulationGrid::getIntMtbOverMIC() const { return _intMtbOverMIC; }
inline double& AntibioticSimulationGrid::getIntMtbOverMIC() { return _intMtbOverMIC; }
inline void AntibioticSimulationGrid::setIntMtbOverMIC(const double& v) {_intMtbOverMIC = v;}

inline const double& AntibioticSimulationGrid::getIntMtbHighRiskOfResistance() const { return _intMtbHighRiskOfResistance; }
inline double& AntibioticSimulationGrid::getIntMtbHighRiskOfResistance() { return _intMtbHighRiskOfResistance; }
inline void AntibioticSimulationGrid::setIntMtbHighRiskOfResistance(const double& v) {_intMtbHighRiskOfResistance = v;}

inline const double& AntibioticSimulationGrid::getExtMtbHighRiskOfResistance() const { return _extMtbHighRiskOfResistance; }
inline double& AntibioticSimulationGrid::getExtMtbHighRiskOfResistance() { return _extMtbHighRiskOfResistance; }
inline void AntibioticSimulationGrid::setExtMtbHighRiskOfResistance(const double& v) {_extMtbHighRiskOfResistance = v;}

inline const double& AntibioticSimulationGrid::getNonrepMtbHighRiskOfResistance() const { return _nonrepMtbHighRiskOfResistance; }
inline double& AntibioticSimulationGrid::getNonrepMtbHighRiskOfResistance() { return _nonrepMtbHighRiskOfResistance; }
inline void AntibioticSimulationGrid::setNonrepMtbHighRiskOfResistance(const double& v) {_nonrepMtbHighRiskOfResistance = v;}

inline const double& AntibioticSimulationGrid::getMacsHighRiskOfResistance() const { return _macsHighRiskOfResistance; }
inline double& AntibioticSimulationGrid::getMacsHighRiskOfResistance() { return _macsHighRiskOfResistance; }
inline void AntibioticSimulationGrid::setMacsHighRiskOfResistance(const double& v) {_macsHighRiskOfResistance = v;}

inline const double& AntibioticSimulationGrid::getCompHighRiskOfResistance() const { return _compHighRiskOfResistance; }
inline double& AntibioticSimulationGrid::getCompHighRiskOfResistance() { return _compHighRiskOfResistance; }
inline void AntibioticSimulationGrid::setCompHighRiskOfResistance(const double& v) {_compHighRiskOfResistance = v;}

inline const double& AntibioticSimulationGrid::getIntMtbSingleDrugKill() const { return _intMtbSingleDrugKill; }
inline double& AntibioticSimulationGrid::getIntMtbSingleDrugKill() { return _intMtbSingleDrugKill; }
inline void AntibioticSimulationGrid::setIntMtbSingleDrugKill(const double &v) {_intMtbSingleDrugKill = v;}
inline void AntibioticSimulationGrid::incIntMtbSingleDrugKill(const double &v) {_intMtbSingleDrugKill = v + _intMtbSingleDrugKill;}

inline const double& AntibioticSimulationGrid::getExtMtbSingleDrugKill() const { return _extMtbSingleDrugKill; }
inline double& AntibioticSimulationGrid::getExtMtbSingleDrugKill() { return _extMtbSingleDrugKill; }
inline void AntibioticSimulationGrid::setExtMtbSingleDrugKill(const double &v) {_extMtbSingleDrugKill = v;}
inline void AntibioticSimulationGrid::incExtMtbSingleDrugKill(const double &v) {_extMtbSingleDrugKill = v + _extMtbSingleDrugKill;}

inline const double& AntibioticSimulationGrid::getNonRepSingleDrugKill() const { return _nonrepSingleDrugKill; }
inline double& AntibioticSimulationGrid::getNonRepSingleDrugKill() {return _nonrepSingleDrugKill; }
inline void AntibioticSimulationGrid::setNonRepSingleDrugKill(const double &v) {_nonrepSingleDrugKill = v;}
inline void AntibioticSimulationGrid::incNonRepSingleDrugKill(const double &v) {_nonrepSingleDrugKill = v + _nonrepSingleDrugKill;}

// Resistance stats
inline DrugMtbStat& AntibioticSimulationGrid::resistant() { return _resistant; }
inline DrugMtbStat& AntibioticSimulationGrid::mdr() { return _mdr; }
inline DrugMtbStat& AntibioticSimulationGrid::compensatedResistant() { return _compensatedResistant; }
inline DrugMtbStat& AntibioticSimulationGrid::mdrCompensatedResistant() { return _mdrCompensatedResistant; }
inline DrugMtbStat& AntibioticSimulationGrid::mdrDoubleCompensatedResistant() { return _mdrDoubleCompensatedResistant; }
inline DrugMtbStat& AntibioticSimulationGrid::cumResistant() { return _cumResistant; }
inline DrugMtbStat& AntibioticSimulationGrid::cumMdr() { return _cumMdr; }
inline DrugMtbStat& AntibioticSimulationGrid::cumCompensated() { return _cumCompensated; }
inline DrugMtbStat& AntibioticSimulationGrid::cumMdrCompensatedResistant() { return _cumMdrCompensatedResistant; }
inline DrugMtbStat& AntibioticSimulationGrid::cumMdrDoubleCompensated() { return _cumMdrDoubleCompensated; }
inline DrugMtbStat& AntibioticSimulationGrid::effluxInduced() { return _effluxInduced; }
inline DrugMtbStat& AntibioticSimulationGrid::doubleEffluxInduced() { return _doubleEffluxInduced; }

// Cumulative concentration stats

inline Scalar& AntibioticSimulationGrid::getDrugConcentration(size_t i) { return _drugConcentration[i]; }
inline Scalar AntibioticSimulationGrid::getDrugConcentration(size_t i) const { return _drugConcentration[i]; }
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentration() const{  return _drugConcentration;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentration() {  return _drugConcentration;}
inline void AntibioticSimulationGrid::setDrugConcentration(size_t i, const Scalar v){  _drugConcentration[i] = v;}

inline Scalar& AntibioticSimulationGrid::getDrugConcentrationNorm(size_t i) {  return _drugConcentrationNorm[i]; }
inline Scalar AntibioticSimulationGrid::getDrugConcentrationNorm(size_t i) const{  return _drugConcentrationNorm[i];}
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationNorm() const{  return _drugConcentrationNorm;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationNorm() {  return _drugConcentrationNorm;}
inline void AntibioticSimulationGrid::setDrugConcentrationNorm(size_t i, const Scalar v){  _drugConcentrationNorm[i] = v;}

inline Scalar& AntibioticSimulationGrid::getDrugConcentrationCaseum(size_t i) {  return _drugConcentrationCaseum[i]; }
inline Scalar AntibioticSimulationGrid::getDrugConcentrationCaseum(size_t i) const{  return _drugConcentrationCaseum[i];}
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationCaseum() const{  return _drugConcentrationCaseum;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationCaseum() {  return _drugConcentrationCaseum;}
inline void AntibioticSimulationGrid::setDrugConcentrationCaseum(size_t i, const Scalar v){  _drugConcentrationCaseum[i] = v;}

inline Scalar& AntibioticSimulationGrid::getDrugConcentrationGran(size_t i){  return _drugConcentrationGran[i];}
inline Scalar AntibioticSimulationGrid::getDrugConcentrationGran(size_t i) const{  return _drugConcentrationGran[i];}
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationGran() const{  return _drugConcentrationGran;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationGran() {  return _drugConcentrationGran;}
inline void AntibioticSimulationGrid::setDrugConcentrationGran(size_t i, const Scalar v){  _drugConcentrationGran[i] = v;}

inline Scalar& AntibioticSimulationGrid::getDrugConcentrationBlood(size_t i){  return _drugConcentrationblood[i];}
inline Scalar AntibioticSimulationGrid::getDrugConcentrationBlood(size_t i) const{  return _drugConcentrationblood[i];}
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationBlood() const{  return _drugConcentrationblood;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationBlood() {  return _drugConcentrationblood;}
inline void AntibioticSimulationGrid::setDrugConcentrationBlood(size_t i, const Scalar v){  _drugConcentrationblood[i] = v;}

inline Scalar& AntibioticSimulationGrid::getDrugConcentrationLung(size_t i){  return _drugConcentrationLung[i];}
inline Scalar AntibioticSimulationGrid::getDrugConcentrationLung(size_t i) const{  return _drugConcentrationLung[i];}
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationLung() const{  return _drugConcentrationLung;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationLung() {  return _drugConcentrationLung;}
inline void AntibioticSimulationGrid::setDrugConcentrationLung(size_t i, const Scalar v){  _drugConcentrationLung[i] = v;}

inline Scalar& AntibioticSimulationGrid::getDrugConcentrationPeriph(size_t i){  return _drugConcentrationPeriph[i];}
inline Scalar AntibioticSimulationGrid::getDrugConcentrationPeriph(size_t i) const{  return _drugConcentrationPeriph[i];}
inline const std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationPeriph() const{  return _drugConcentrationPeriph;}
inline std::vector<Scalar>& AntibioticSimulationGrid::getDrugConcentrationPeriph() {  return _drugConcentrationPeriph;}
inline void AntibioticSimulationGrid::setDrugConcentrationPeriph(size_t i, const Scalar v){  _drugConcentrationPeriph[i] = v;}

inline GrAccumulator<double>& AntibioticSimulationGrid::getdrugDistStat() { return  _drugDistStat;}
inline const GrAccumulator<double>& AntibioticSimulationGrid::getdrugDistStat() const {  return _drugDistStat;}

// END Stats get set functions

// Parameter get functions
inline const int& AntibioticSimulationGrid::PKModel() const { return _PKModel; }
inline int& AntibioticSimulationGrid::PKModel() { return _PKModel; };

inline const double& AntibioticSimulationGrid::vesselActivity() const { return _vesselActivity; }
inline double& AntibioticSimulationGrid::vesselActivity() { return _vesselActivity; }

inline const int& AntibioticSimulationGrid::partitionUsingSteadyState() const { return _partitionUsingSteadyState; }
inline int& AntibioticSimulationGrid::partitionUsingSteadyState() { return _partitionUsingSteadyState; }

inline const double& AntibioticSimulationGrid::degRateConst() const { return _degRateConst; }
inline double& AntibioticSimulationGrid::degRateConst() { return _degRateConst; }

inline const double& AntibioticSimulationGrid::degRateConstInt() const { return _degRateConstInt; }
inline double& AntibioticSimulationGrid::degRateConstInt() { return _degRateConstInt; }

inline const double& AntibioticSimulationGrid::diffusivity() const { return _diffusivity; }
inline double& AntibioticSimulationGrid::diffusivity() { return _diffusivity; }

inline const double& AntibioticSimulationGrid::minDiffusivity() const { return _minDiffusivity; }
inline double& AntibioticSimulationGrid::minDiffusivity() { return _minDiffusivity; }

inline const double& AntibioticSimulationGrid::kAbs() const { return _kAbs; }
inline double& AntibioticSimulationGrid::kAbs() { return _kAbs; }

inline const double& AntibioticSimulationGrid::q() const { return _q; }
inline double& AntibioticSimulationGrid::q() { return _q; }

inline const double& AntibioticSimulationGrid::volDistCentral() const { return _volDistCentral; }
inline double& AntibioticSimulationGrid::volDistCentral() { return _volDistCentral; }

inline const double& AntibioticSimulationGrid::volDistPeriph() const { return _volDistPeriph; }
inline double& AntibioticSimulationGrid::volDistPeriph() { return _volDistPeriph; }

inline const double& AntibioticSimulationGrid::clearance() const { return _clearance; }
inline double& AntibioticSimulationGrid::clearance() { return _clearance; }

inline const double& AntibioticSimulationGrid::dose() const { return _dose; }
inline double& AntibioticSimulationGrid::dose() { return _dose; }

inline const int& AntibioticSimulationGrid::dosageInterval() const { return _dosageInterval; }
inline int &AntibioticSimulationGrid::dosageInterval() { return _dosageInterval; }

inline const double& AntibioticSimulationGrid::cellUptake() const { return _cellUptake; }
inline double& AntibioticSimulationGrid::cellUptake() { return _cellUptake; }

inline const double& AntibioticSimulationGrid::vascularPermeability() const { return _vascularPermeability; }
inline double& AntibioticSimulationGrid::vascularPermeability() { return _vascularPermeability; }

inline const double& AntibioticSimulationGrid::permCoeff() const { return _permCoeff; }
inline double& AntibioticSimulationGrid::permCoeff() { return _permCoeff; }

inline const double& AntibioticSimulationGrid::intraActivity() const { return _intraActivity; }
inline double& AntibioticSimulationGrid::intraActivity() { return _intraActivity; }

inline const double& AntibioticSimulationGrid::extraActivity() const { return _extraActivity; }
inline double& AntibioticSimulationGrid::extraActivity() { return _extraActivity; }

inline const double& AntibioticSimulationGrid::nonrepActivity() const { return _nonrepActivity; }
inline double& AntibioticSimulationGrid::nonrepActivity() { return _nonrepActivity; }

inline const double& AntibioticSimulationGrid::c50Extra() const { return _c50Extra; }
inline double& AntibioticSimulationGrid::c50Extra() { return _c50Extra; }

inline const double& AntibioticSimulationGrid::c50Intra() const { return _c50Intra; }
inline double& AntibioticSimulationGrid::c50Intra() { return _c50Intra; }

inline const double& AntibioticSimulationGrid::c50Nonrep() const { return _c50Nonrep; }
inline double& AntibioticSimulationGrid::c50Nonrep() { return _c50Nonrep; }

inline const double& AntibioticSimulationGrid::consumptionRate() const { return _consumptionRate; }
inline double& AntibioticSimulationGrid::consumptionRate() { return _consumptionRate; }

inline const double& AntibioticSimulationGrid::consumptionHalfSat() const { return _consumptionHalfSat; }
inline double& AntibioticSimulationGrid::consumptionHalfSat() { return _consumptionHalfSat; }

inline const double& AntibioticSimulationGrid::mic() const { return _mic; }
inline double& AntibioticSimulationGrid::mic() { return _mic; }

inline const double& AntibioticSimulationGrid::hillExtra() const { return _hillExtra; }
inline double& AntibioticSimulationGrid::hillExtra() { return _hillExtra; }

inline const double& AntibioticSimulationGrid::hillIntra() const { return _hillIntra; }
inline double& AntibioticSimulationGrid::hillIntra() { return _hillIntra; }

inline const double& AntibioticSimulationGrid::hillNonrep() const { return _hillNonrep; }
inline double& AntibioticSimulationGrid::hillNonrep() { return _hillNonrep; }

inline const double& AntibioticSimulationGrid::caseumUnboundFraction() const { return _caseumUnboundFraction; }
inline double& AntibioticSimulationGrid::caseumUnboundFraction() { return _caseumUnboundFraction; }

inline const double& AntibioticSimulationGrid::caseumBindingRateConst() const { return _caseumBindingRateConst; }
inline double& AntibioticSimulationGrid::caseumBindingRateConst() { return _caseumBindingRateConst; }

inline const double& AntibioticSimulationGrid::epitheliumBindingKa() const { return _epitheliumBindingKa; }
inline double& AntibioticSimulationGrid::epitheliumBindingKa() { return _epitheliumBindingKa; }

inline const double& AntibioticSimulationGrid::epitheliumBindingRateConst() const { return _epitheliumBindingRateConst; }
inline double& AntibioticSimulationGrid::epitheliumBindingRateConst() { return _epitheliumBindingRateConst; }

inline const double& AntibioticSimulationGrid::cellularKout() const { return _cellularKout; }
inline double& AntibioticSimulationGrid::cellularKout() { return _cellularKout; }

inline const double& AntibioticSimulationGrid::MW() const { return _MW; }
inline double& AntibioticSimulationGrid::MW() { return _MW; }

inline const int& AntibioticSimulationGrid::resistRiskIndicatorFlag() const { return _resistRiskIndicatorFlag; }
inline int& AntibioticSimulationGrid::resistRiskIndicatorFlag() { return _resistRiskIndicatorFlag; }

inline const double& AntibioticSimulationGrid::numberResistMutations() const { return _numberResistMutations; }
inline double& AntibioticSimulationGrid::numberResistMutations() { return _numberResistMutations; }

inline const double& AntibioticSimulationGrid::relativeFitnessResist() const { return _relativeFitnessResist; }
inline double& AntibioticSimulationGrid::relativeFitnessResist() { return _relativeFitnessResist; }

inline const double& AntibioticSimulationGrid::numberCompensatoryMutations() const { return _numberCompensatoryMutations; }
inline double& AntibioticSimulationGrid::numberCompensatoryMutations() { return _numberCompensatoryMutations; }

inline const double& AntibioticSimulationGrid::foldIncreaseC50FromResistance() const {return _foldIncreaseC50FromResistance; }
inline double& AntibioticSimulationGrid::foldIncreaseC50FromResistance() { return _foldIncreaseC50FromResistance;}

inline const int& AntibioticSimulationGrid::numTimeStepsEffluxInduction() const {return _numTimeStepsEffluxInduction; }
inline int &AntibioticSimulationGrid::numTimeStepsEffluxInduction() { return _numTimeStepsEffluxInduction;}

inline const double& AntibioticSimulationGrid::minEffluxInduction() const {return _minEffluxInduction; }
inline double& AntibioticSimulationGrid::minEffluxInduction() { return _minEffluxInduction;}

inline const double& AntibioticSimulationGrid::maxEffluxInduction() const {return _maxEffluxInduction; }
inline double& AntibioticSimulationGrid::maxEffluxInduction() { return _maxEffluxInduction;}

inline const double& AntibioticSimulationGrid::concentrationChangeByEfflux() const {return _concentrationChangeByEfflux; }
inline double& AntibioticSimulationGrid::concentrationChangeByEfflux() { return _concentrationChangeByEfflux;}

inline const int& AntibioticSimulationGrid::numberOfRandomMonoResistantBugsToAdd() const {return _numberOfRandomMonoResistantBugsToAdd; }
inline int& AntibioticSimulationGrid::numberOfRandomMonoResistantBugsToAdd() { return _numberOfRandomMonoResistantBugsToAdd;}

inline const double& AntibioticSimulationGrid::volDistLung() const {return _volDistLung; }
inline double& AntibioticSimulationGrid::volDistLung() { return _volDistLung;}

// Calculate MIC using the Emax model to get the concentration where fractionKill% of max killing is achieved
inline double AntibioticSimulationGrid::calculateMicFromC50(double c50, double hill, double fractionKill)
{
    return pow((pow(c50,hill)*fractionKill)/(1-fractionKill),1/hill);
}

// END parameter get functions

//Partitioning ODES
struct GridAntibioticPartitioningFunc : ODESolvers::DerivativeFunc
{
    struct Params_t
    {
        double* C;
        double* kfc;
        double* krc;
        double* kfe;
        double* kre;
        double* kin;
        double* kout;
    };
    GridAntibioticPartitioningFunc() : DerivativeFunc() {}
    virtual ~GridAntibioticPartitioningFunc() {}
    /*virtual*/ void operator()(const ODESolvers::ODEState& vecread, double t, ODESolvers::Derivative& vecwrite, void* params) const;

};

inline GridAntibioticPartitioningFunc* AntibioticSimulationGrid::buildDerivFunc()
{
  GridAntibioticPartitioningFunc* d = new GridAntibioticPartitioningFunc();
  return d;
}


// Resistance risk
// Use the flag to determine which indicator to use (0=AUC; 1=Cmax).
// Pass back the intracellular and extracellualar values of the appropriate indicator
// Also pass back the appropriate lower and upper bounds (we have to adjust 24 hour AUC or t>MIC
// bounds for dosing intervals. Since Cmax is not cumulative it does not need adjustment)

// INH uses AUC as an indicator - data for INH is not as good.
// RIF uses Cmax as an indicator of drug resistance risk. According to Gumbo et al 2007, AAC 51(11):3781.

inline void AntibioticSimulationGrid::getResistRiskValues(int riskFlag, double& riskIndicatorValueIntra, double& riskIndicatorValueExtra,
                         double& riskLowerBound, double& riskUpperBound, const Pos& p)
{
    if (riskFlag == 0)
    {
        riskIndicatorValueIntra = aucIntraGrid(p);
        riskIndicatorValueExtra = aucExtraGrid(p);
        riskLowerBound = _aucRiskLB / TIME_STEPS_PER_DAY * _PARAM(Drugs_drugConcentrationStatInterval);
        riskUpperBound = _aucRiskUB / TIME_STEPS_PER_DAY * _PARAM(Drugs_drugConcentrationStatInterval);
    }
    else if (riskFlag == 1)
    {
        riskIndicatorValueIntra = cmaxIntraGrid(p);
        riskIndicatorValueExtra = cmaxExtraGrid(p);
        riskLowerBound = _cmaxRiskLB;
        riskUpperBound = _cmaxRiskUB;
    }
}

// END resistance risk

inline double AntibioticSimulationGrid::computeInteractEffectiveConc(double fic,
                                            const vector<double>& concRatio)
{
    double effectiveConc = 0.0;
    for(size_t i = 0; i < concRatio.size(); ++i)
    {
        effectiveConc += pow(concRatio[i], fic);
    }

    effectiveConc = pow(effectiveConc, 1.0 / fic);

    return effectiveConc;
}

inline double AntibioticSimulationGrid::computeInteractAdjConc(double fic,
                                                               const vector<double>& concVec,
                                                               const vector<double>& emaxVec,
                                                               const vector<double>& c50Vec,
                                                               const vector<double>& hVec,
                                                               double intc50, double intEmax, double intH)
{
    double effectiveConc = 0.0;
    double adjConc;
    for(size_t i = 0; i < concVec.size(); ++i)
    {
        adjConc = 0.0;
        double c50m = pow(intc50, intH);
        double c = pow(concVec[i], hVec[i]);
        double c50i = pow(c50Vec[i], hVec[i]);
        adjConc = (c50m * c) / ((intEmax/emaxVec[i])*(c50i + c) - c);
        adjConc = pow(adjConc, 1.0 / intH);
        effectiveConc += pow(adjConc, fic);
    }

    effectiveConc = pow(effectiveConc, 1.0 / fic);

    return effectiveConc;
}

inline bool AntibioticSimulationGrid::checkShouldKill(double killRate)
{
    bool shouldKill = false;

    // Calculate killing probability from rate.
    // Do not multiply by time because units are timesteps i.e. * 1.0
    double killProb = 1 - exp(-killRate);

    if (g_Rand.getReal() < killProb)
    {
        shouldKill = true;
    }
    return shouldKill;
}

inline double AntibioticSimulationGrid::getKillThreshold(double percThreshold) const
{
    double threshold = percThreshold * _mic;
    return threshold;
}



template<class Archive>
void AntibioticSimulationGrid::serialize(Archive& ar, const unsigned int /*version*/)
{
    ar & BOOST_SERIALIZATION_NVP(_dim);
    ar & BOOST_SERIALIZATION_NVP(_name);
    ar & BOOST_SERIALIZATION_NVP(_PKModel);
    ar & BOOST_SERIALIZATION_NVP(_cutOffValue);

    // Parameters
    ar & BOOST_SERIALIZATION_NVP(_vesselActivity);
    ar & BOOST_SERIALIZATION_NVP(_partitionUsingSteadyState);
    ar & BOOST_SERIALIZATION_NVP(_degRateConst);
    ar & BOOST_SERIALIZATION_NVP(_degRateConstInt);
    ar & BOOST_SERIALIZATION_NVP(_diffusivity);
    ar & BOOST_SERIALIZATION_NVP(_minDiffusivity);
    ar & BOOST_SERIALIZATION_NVP(_kAbs);
    ar & BOOST_SERIALIZATION_NVP(_q);
    ar & BOOST_SERIALIZATION_NVP(_volDistCentral);
    ar & BOOST_SERIALIZATION_NVP(_volDistPeriph);
    ar & BOOST_SERIALIZATION_NVP(_clearance);
    ar & BOOST_SERIALIZATION_NVP(_dose);
    ar & BOOST_SERIALIZATION_NVP(_dosageInterval);
    ar & BOOST_SERIALIZATION_NVP(_cellUptake);
    ar & BOOST_SERIALIZATION_NVP(_vascularPermeability);
    ar & BOOST_SERIALIZATION_NVP(_permCoeff);
    ar & BOOST_SERIALIZATION_NVP(_intraActivity);
    ar & BOOST_SERIALIZATION_NVP(_extraActivity);
    ar & BOOST_SERIALIZATION_NVP(_nonrepActivity);
    ar & BOOST_SERIALIZATION_NVP(_c50Extra);
    ar & BOOST_SERIALIZATION_NVP(_c50Intra);
    ar & BOOST_SERIALIZATION_NVP(_c50Nonrep);
    ar & BOOST_SERIALIZATION_NVP(_consumptionRate);
    ar & BOOST_SERIALIZATION_NVP(_consumptionHalfSat);
    ar & BOOST_SERIALIZATION_NVP(_mic);
    ar & BOOST_SERIALIZATION_NVP(_hillExtra);
    ar & BOOST_SERIALIZATION_NVP(_hillIntra);
    ar & BOOST_SERIALIZATION_NVP(_hillNonrep);
    ar & BOOST_SERIALIZATION_NVP(_caseumUnboundFraction);
    ar & BOOST_SERIALIZATION_NVP(_caseumBindingRateConst);
    ar & BOOST_SERIALIZATION_NVP(_epitheliumBindingKa);
    ar & BOOST_SERIALIZATION_NVP(_epitheliumBindingRateConst);
    ar & BOOST_SERIALIZATION_NVP(_cellularKout);
    ar & BOOST_SERIALIZATION_NVP(_MW);
    ar & BOOST_SERIALIZATION_NVP(_resistRiskIndicatorFlag);
    ar & BOOST_SERIALIZATION_NVP(_aucRiskLB);
    ar & BOOST_SERIALIZATION_NVP(_aucRiskUB);
    ar & BOOST_SERIALIZATION_NVP(_cmaxRiskLB);
    ar & BOOST_SERIALIZATION_NVP(_cmaxRiskUB);
    ar & BOOST_SERIALIZATION_NVP(_numberResistMutations);
    ar & BOOST_SERIALIZATION_NVP(_relativeFitnessResist);
    ar & BOOST_SERIALIZATION_NVP(_numberCompensatoryMutations);
    ar & BOOST_SERIALIZATION_NVP(_foldIncreaseC50FromResistance);
    ar & BOOST_SERIALIZATION_NVP(_numTimeStepsEffluxInduction);
    ar & BOOST_SERIALIZATION_NVP(_minEffluxInduction);
    ar & BOOST_SERIALIZATION_NVP(_maxEffluxInduction);
    ar & BOOST_SERIALIZATION_NVP(_concentrationChangeByEfflux);
    ar & BOOST_SERIALIZATION_NVP(_numberOfRandomMonoResistantBugsToAdd);
    ar & BOOST_SERIALIZATION_NVP(_volDistLung);

    // Concentration Grids
    ar & BOOST_SERIALIZATION_NVP(_pMat);

    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    ar & BOOST_SERIALIZATION_NVP(_cnadiChemical);
    #endif

    ar & BOOST_SERIALIZATION_NVP(_currentGrid);
    ar & BOOST_SERIALIZATION_NVP(_nextGrid);

    // Vascular variables
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentration);
    ar & BOOST_SERIALIZATION_NVP(_absComp1Concentration);
    ar & BOOST_SERIALIZATION_NVP(_absComp2Concentration);
    ar & BOOST_SERIALIZATION_NVP(_perCompConcentration);
    ar & BOOST_SERIALIZATION_NVP(_lungCompConcentration);
    ar & BOOST_SERIALIZATION_NVP(_intMacConcentration);

    ar & BOOST_SERIALIZATION_NVP(_initvector);

    // Stats Grids
    ar & BOOST_SERIALIZATION_NVP(_intracellularGrid);
    ar & BOOST_SERIALIZATION_NVP(_totalGrid);
    ar & BOOST_SERIALIZATION_NVP(_aucIntraGrid);
    ar & BOOST_SERIALIZATION_NVP(_aucExtraGrid);
    ar & BOOST_SERIALIZATION_NVP(_aucTotalGrid);
    ar & BOOST_SERIALIZATION_NVP(_cmaxIntraGrid);
    ar & BOOST_SERIALIZATION_NVP(_cmaxExtraGrid);
    ar & BOOST_SERIALIZATION_NVP(_cmaxTotalGrid);


    // Stats
    ar & BOOST_SERIALIZATION_NVP(_totalWholeGrid);
    ar & BOOST_SERIALIZATION_NVP(_totalGran);
    ar & BOOST_SERIALIZATION_NVP(_totalNorm);
    ar & BOOST_SERIALIZATION_NVP(_totalCaseum);
    ar & BOOST_SERIALIZATION_NVP(_partialCaseum);
    ar & BOOST_SERIALIZATION_NVP(_avgConcInfInt);
    ar & BOOST_SERIALIZATION_NVP(_avgConcExtMtb);
    ar & BOOST_SERIALIZATION_NVP(_avgConcNonMtb);

    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationArea);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationAreaNorm);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationAreaCaseum);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationAreaGran);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationAreaBlood);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationAreaLung);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationAreaPeriph);
    ar & BOOST_SERIALIZATION_NVP(_cmax);
    ar & BOOST_SERIALIZATION_NVP(_cmaxNorm);
    ar & BOOST_SERIALIZATION_NVP(_cmaxCaseum);
    ar & BOOST_SERIALIZATION_NVP(_cmaxGran);
    ar & BOOST_SERIALIZATION_NVP(_cmaxBlood);
    ar & BOOST_SERIALIZATION_NVP(_compOverMIC);
    ar & BOOST_SERIALIZATION_NVP(_extMtbOverMIC);
    ar & BOOST_SERIALIZATION_NVP(_macsOverMIC);
    ar & BOOST_SERIALIZATION_NVP(_intMtbOverMIC);
    ar & BOOST_SERIALIZATION_NVP(_intMtbHighRiskOfResistance);
    ar & BOOST_SERIALIZATION_NVP(_extMtbHighRiskOfResistance);
    ar & BOOST_SERIALIZATION_NVP(_nonrepMtbHighRiskOfResistance);
    ar & BOOST_SERIALIZATION_NVP(_macsHighRiskOfResistance);
    ar & BOOST_SERIALIZATION_NVP(_compHighRiskOfResistance);
    ar & BOOST_SERIALIZATION_NVP(_intMtbSingleDrugKill);
    ar & BOOST_SERIALIZATION_NVP(_extMtbSingleDrugKill);
    ar & BOOST_SERIALIZATION_NVP(_nonrepSingleDrugKill);

    // Resistance stats
    ar & BOOST_SERIALIZATION_NVP(_resistant);
    ar & BOOST_SERIALIZATION_NVP(_mdr);
    ar & BOOST_SERIALIZATION_NVP(_compensatedResistant);
    ar & BOOST_SERIALIZATION_NVP(_mdrCompensatedResistant);
    ar & BOOST_SERIALIZATION_NVP(_mdrDoubleCompensatedResistant);
    ar & BOOST_SERIALIZATION_NVP(_cumResistant);
    ar & BOOST_SERIALIZATION_NVP(_cumMdr);
    ar & BOOST_SERIALIZATION_NVP(_cumCompensated);
    ar & BOOST_SERIALIZATION_NVP(_cumMdrCompensatedResistant);
    ar & BOOST_SERIALIZATION_NVP(_cumMdrDoubleCompensated);
    ar & BOOST_SERIALIZATION_NVP(_effluxInduced);
    ar & BOOST_SERIALIZATION_NVP(_doubleEffluxInduced);

    ar & BOOST_SERIALIZATION_NVP(_drugConcentration);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationNorm);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationCaseum);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationGran);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationblood);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationLung);
    ar & BOOST_SERIALIZATION_NVP(_drugConcentrationPeriph);

    ar & BOOST_SERIALIZATION_NVP(_drugDistStat);
}

#endif /* ANTIBIOTICSIMULATIONGRID_H */
