
#include "drug-kill-indigo.h"
#include "grgrid.h"
#include "stat.h"
#include "antibioticconcentration.h"
#include <map>

BOOST_CLASS_EXPORT_IMPLEMENT(DrugKillIndigo)

// Set to 1 for some standard debug output.
#define DBGDKI 0

// Needed for serialization.
DrugKillIndigo::DrugKillIndigo()
    : DrugKill(DRIM_NONE)
    , _drugInteractionList()
{
    _drugInteractionMethod = DRIM_INDIGO;
}

DrugKillIndigo::DrugKillIndigo(const string& drugInteractionFileName,
                               const std::map<string, int>& drugMap)
    : DrugKill(DRIM_INDIGO)
    , _drugInteractionList()
{
    #if DBGDKI
    cout << "DrugKillIndigo::DrugKillIndigo " << this << endl;
    #endif

    createFicTable(drugInteractionFileName, drugMap);
    
    #if DBGDKI
    cout << "DrugKillIndigo::DrugKillIndigo _drugInteractionList.size(): "
         << _drugInteractionList.size() << endl;
    #endif
}

DrugKillIndigo::~DrugKillIndigo()
{
    #if DBGDKI
    cout << "DrugKillIndigo::~DrugKillIndigo" << endl;
    #endif
}

// Read the drug interaction file.
void DrugKillIndigo::createFicTable(const string& drugInteractionFileName,
                                  const std::map<string, int>& drugMap)
{
    
    #if DBGDKI
    cout << "DrugKillIndigo::createFicTable start" << endl;
    #endif
    // Make the drug interaction list big enough to hold all possible
    // combinations of the drugs being used in the simulation. Some of those
    // combintations might not be specified in the drug interaction file, but
    // we still must make room for them. Also element 0 will be for the drug
    // combination with no drugs. This won't occur - no drug killing will take
    // place in that case, so no FIC will be needed for it (or make sense for
    // it), but again, a slot is needed for it.  If there are 3 drugs there are
    // 8 combinations, "000", "001", "010", "011" "100", "101", "110", "111". 8
    // is 2^3 = 1 << 3.
    int drugInteractionSize = 1 << drugMap.size();
    _drugInteractionList.resize(drugInteractionSize);

    // Assign entries that won't be in the drug interaction file.
    initDrugInteractionTable(drugMap);

    std::ifstream inFile(drugInteractionFileName.c_str());

    if(!inFile.good())
    {
        std::cerr << "failed to open drug interaction file '"
             << drugInteractionFileName << "'."
             << std::endl;
        throw std::runtime_error("Fatal error");
    }
    
    #if DBGDKI
        cout << "after if(!inFile.good())" << endl;
        cout << "inFile.peek() '" << inFile.peek() << "'." << endl;
        cout << "char_traits<char>::eof() '" << char_traits<char>::eof()
             << "'." << endl;
    #endif

    string line;
    int lineCount = 0;
    bool errorsFound = false;

    // Using peek instead of !inFile.eof() avoids reading a blank line at the
    // end of the file.  The reasons are unknown.
    while(inFile.peek() != char_traits<char>::eof())
    {
        std::getline(inFile, line);
        ++lineCount;

        #if DBGDKI
            cout << lineCount << ": " << line << endl;
        #endif

        // Get the drug names and FIC from the line.
        errorsFound |= parseDrugInteractionLine(line, lineCount, drugMap);
    }

    inFile.close();

    if(lineCount == 0)
    {
        cout << "No lines read from drug interaction file '"
             << drugInteractionFileName << "'." << endl;
        throw std::runtime_error("Fatal error");
    }

    // No longer do this as the latest drug interaction file does not have all
    // combinations of the drugs included in the file.
    //errorsFound |= checkDrugInteractionTable(drugInteractionList);

    if(errorsFound)
    {
        throw std::runtime_error("Fatal error");
    }
}

// Initialize the drug interaction table with entries that won't be in the drug
// interaction file.
//
// The drug interaction file does not have entries for a single drug by itself.
// Make those have an FIC of 1. The single drug entries are those with an index
// that is a power of 2.
//
// Also define an entry for no drugs, with an FIC of -1.
void DrugKillIndigo::initDrugInteractionTable(const std::map<string, int>& drugMap)
{
    _drugInteractionList[0] = DrugInteraction("", -1.0);
    for(std::map<string, int>::const_iterator it = drugMap.begin();
        it != drugMap.end(); ++it)
    {
        string names = it->first;
        int idx = it->second;
        _drugInteractionList[idx] = DrugInteraction(names, 1.0);
    }
}

// For a line read from the drug interaction file, parse it and define the
// index into the drug interaction table based on the drugs for that line. Use
// the index to define a DrugInteraction object for that table entry, to hold
// the drugs names and the FIC value.
bool DrugKillIndigo::parseDrugInteractionLine(const string& line, int lineCount,
                                              const std::map<string, int>& drugMap)
{
    boost::regex reCsv(reCsvStr);
    boost::sregex_token_iterator endToken;
    boost::sregex_token_iterator tokenIterator(line.begin(), line.end(),
                                               reCsv, -1);
    string names;
    int idx = 0;
    int fieldCount = 0;
    int expectedFieldCount = 7;
    bool skipped = false;
    bool errorsFound = false;
    for(fieldCount = 0; tokenIterator != endToken; ++tokenIterator)
    {
        string token = *tokenIterator;
        ++fieldCount;
        if(fieldCount < expectedFieldCount)
        {
            // Process a drug name.
            if(token.empty())
            {
                // Skip an empty drug.
                #if DBGDKI
                    cout << "Skipping empty drug" << endl;
                #endif
                continue;
            }

            if(drugMap.count(token) == 0)
            {
                // Skip this line because it has a drug that is not being
                // used in this run.
                skipped = true;
                #if DBGDKI
                    cout << "Skipping drug not being used '" << token << "'" << endl;
                #endif
                break;
            }

            names += (fieldCount > 1 ? "," : "") + token;
            idx += drugMap.at(token);
            #if DBGDKI
                cout << "Using drug '" << token << "' names: '" << names
                     << "' idx: " << idx << endl;
            #endif
        }
        else if(fieldCount == expectedFieldCount)
        {
            // Process FIC value.
            if(names.empty())
            {
                errorsFound = true;
                std::cerr << "Drug interaction file line " << lineCount
                          << " with fic " << token << " has no drug names."
                          << std::endl;
            }

            double fic = 0.0;
            std::istringstream num(token);
            num >> fic;
            // eof check to avoid flagging something like "123Z" as valid.
            if(num.fail() || !num.eof())
            {
                errorsFound = true;
                std::cerr << "Drug interaction file line " << lineCount
                          << " has an fic that is not a valid number: '"
                          << token << "'." << std::endl;
                break;
            }

            if(fic < 0)
            {
                errorsFound = true;
                std::cerr << "Drug interaction file line " << lineCount
                          << " has an fic < 0 of " << fic << "."
                          << std::endl;
                break;
            }

            #if DBGDKI
                cout << "fic: " << fic << " names: '" << names
                     << "' idx: " << idx << endl;
            #endif
            _drugInteractionList[idx] = DrugInteraction(names, fic); 
        }
        else
        {
            // ERROR: an extra field.
            std::cerr << "Extra field '" << token << "' was read from line "
                      << lineCount << std::endl;
            throw std::runtime_error("Fatal error");
        }
    }

    if(!skipped && fieldCount < expectedFieldCount)
    {
        std::cerr << "Too few fields read from drug interaction file line "
                  << lineCount << ", 7 required, " << fieldCount
                  << " read."
                  << std::endl;
        errorsFound = true;
    }

    return errorsFound;
}

// Check that the drug interaction entire table is correct.
//
// Entry 0 should have empty names and fic of -1.
//
// All other entries should have a non-empty names and fic >= 0.
//
// This may duplicate some errors found when creating the table, but will also
// check for table entries that were never assigned values from the drug
// interaction file.
//
// This doesn't check for duplicate lines in the drug interaction file - lines
// with the same set of drug names (not necessarily the same fic).
bool DrugKillIndigo::checkDrugInteractionTable() const
{

    cout << endl;
    int badCount = 0;
    for(size_t i = 0; i < _drugInteractionList.size(); ++i)
    {
        string names = _drugInteractionList[i]._names;
        double fic = _drugInteractionList[i]._fic;
        
        if(i == 0)
        {
            if(!names.empty() || fic != -1)
            {
                ++badCount;
                if(!names.empty())
                {
                    std::cerr << "Invalid drug interaction table entry 0,"
                              << " for no drugs, has non-empty drug names of '"
                              << names << "'." << std::endl;
                }
    
                if(fic != -1)
                {
                    std::cerr << "Invalid drug interaction table entry 0,"
                              << " for no drugs, has an fic other than -1 of "
                              << fic << std::endl;
                }
            }
        }
        else if(names.empty() || fic < 0)
        {
            ++badCount;
            if(names.empty())
            {
                std::cerr << "Invalid drug interaction table entry "
                          << i << " has no names. " << std::endl;
            }

            if(fic < 0)
            {
                std::cerr << "Invalid drug interaction table entry "
                          << i << " has an fic < 0 of " << fic
                          << ", with drug names '" << names << "'."
                          << std::endl;
            }

            std::cerr << std::endl
                      << "This is usually due to the drug interaction file not"
                      << " having an entry for a combination of the drugs"
                      << " being used in the simulation."
                      << std::endl
                      << "This is especially true if the file only has entries"
                      << " for up to N combinations of drugs but M > N drugs"
                      << " are being used,"
                      << std::endl
                      << "ex. if the file only has entries for up to 4"
                      << " combinations of drugs, but 5 drugs are being used."
                      << std::endl;
        }
    }

    bool errorsFound = false;
    if(badCount > 0)
    {
        errorsFound = true;
        std::cerr << "There were " << badCount
                  << " bad drug interaction table entries."
                  << std::endl;
    }

    return errorsFound;
}

// Check whether an extra-cellular mtb should die from drugs.
bool DrugKillIndigo::checkKillExtra(const AntibioticConcentrationList& mtbAcList,
                                    MtbGroupLocation location, GrGrid& pg,
                                    const Pos& pos, double& interactFactor)
{
    #if DBGDKI
    cout << "DrugKillIndigo::checkKillExtra start" << endl;
    #endif

    bool shouldKill = false;
    int ficIndex = 0;
    double fic = 0.0;
    double intEmax;
    double intc50;
    double intH;
    vector<double> concVec;
    vector<double> emaxVec;
    vector<double> c50Vec;
    vector<double> hVec;

    getDrugExtraInteractInfoDoseAdj(location, pos, mtbAcList, concVec,
                                    emaxVec, c50Vec, hVec, intEmax, intc50,
                                    intH, ficIndex, pg);

    
    #if DBGDKI
        cout << "extra ficIndex: " << ficIndex  << " names: '"
             << _drugInteractionList[ficIndex]._names << "' fic: "
             << _drugInteractionList[ficIndex]._fic << endl;
    #endif
    if(ficIndex > 0)
    {
        shouldKill = checkKillInteractAdjDose("Extra-cellular", ficIndex,
                                              concVec, emaxVec, c50Vec, hVec,
                                              intc50, intEmax, intH,
                                              interactFactor);
    }
    else
    {
        // No drugs met the conditions for drug interaction killing, so use the
        // single drug killing.
        shouldKill = DrugKill::checkKillExtra(mtbAcList, location, pg, pos, interactFactor);
        interactFactor += 1.0;
    }

    pg.setExtFIC(pos, fic);

    return shouldKill;
}

void DrugKillIndigo::updateStatsExtra(Stats& stats, MtbGroupLocation location,
                                      double mtbLiveCount)
{
    updateStatsExtraNonSingle(stats, location, mtbLiveCount);
}

void DrugKillIndigo::updateInteractFactorStatExtra(Stats& stats,
                                                   MtbGroupLocation location,
                                                   double interactFactor)
{
    updateInteractFactorStatExtraNonSingle(stats, location, interactFactor);
}

// Get the drug interaction information for extra-cellular drug killing,
// using the dose/adjustment concentration model
//
// An ficIndex into the drug interaction table is defined for the set of drugs
// that meet the killing threshold. If no drugs meet the threshold then it is
// 0.
void DrugKillIndigo::getDrugExtraInteractInfoDoseAdj(
                         MtbGroupLocation location, const Pos& pos,
                         const AntibioticConcentrationList& mtbAcList,
                         vector<double>& adjConc, vector<double>& emaxVec,
                         vector<double>& c50Vec, vector<double>& hVec,
                         double& intEmax, double& intc50, double& intH,
                         int& ficIndex, const GrGrid& pg)
{
    double c50 = 0.0;
    double emax = 0.0;
    double h = 0.0;
    double conc = 0.0;
    double effectiveConc = 0.0;
    intEmax = 0.0;
    intc50 = 0.0;
    intH = 0.0;
    ficIndex = 0.0;
    int count = 0;
    for (unsigned abxInd=0; abxInd < mtbAcList.size(); abxInd++)
    {
        AntibioticConcentration* pMtbAbxConc = mtbAcList.at(abxInd);
        AntibioticSimulationGrid* pAsg = pMtbAbxConc->antibioticSimulationGrid();

        conc = getTotalConc(pMtbAbxConc, pg, pos, abxInd);
        effectiveConc = pMtbAbxConc->effectiveConcentration(conc);
        if(effectiveConc < pMtbAbxConc->getKillThreshold(_PARAM(Drugs_interactThreshold)))
        {
            continue;
        }

        if (location == NON_REPLICATING_EXTRA_CELLULAR)
        {
            c50 = pMtbAbxConc->mtbSpecificC50Nonrep();
            emax = pAsg->nonrepActivity();
            h = pAsg->hillNonrep();
        }
        else
        {
            c50 = pMtbAbxConc->mtbSpecificC50Extra();
            emax = pAsg->extraActivity();
            h = pAsg->hillExtra();
        }

        // Get all relevant PD information for rate constant
        // calculation
        adjConc.push_back(effectiveConc);
        c50Vec.push_back(c50);
        emaxVec.push_back(emax);
        hVec.push_back(h);

        // check to see if the current Emax is larger,
        // if it is, store that as the interaction Emax,
        // and also store that antbiotic's corresponding c50 and h
        if(emax > intEmax){
            intEmax = emax;
            intc50 = c50;
            intH = h;
        }

        // Set the bit corresponding to the drug.
        ficIndex += (1 << count);
    }
}

// Check whether an intra-cellular mtb should die from drugs.
bool DrugKillIndigo::checkKillIntra(
              const AntibioticConcentrationList& mtbAcList,
              const AntibioticConcentrationList& immuneCellAcList,
              GrGrid& pg, const Pos& pos, double& interactFactor)
{
    #if DBGDKI
    cout << "DrugKillIndigo::checkKillIntra start" << endl;
    #endif

    bool shouldKill = false;
    int ficIndex = 0;
    double fic = 0.0;
    double intEmax = 0.0;
    double intc50;
    double intH;
    vector<double> concVec;
    vector<double> emaxVec;
    vector<double> c50Vec;
    vector<double> hVec;

    getDrugIntraInteractInfoDoseAdj(mtbAcList, immuneCellAcList, concVec,
                                    emaxVec, c50Vec, hVec, intEmax, intc50,
                                    intH, ficIndex, pg ,pos);

    #if DBGDKI
        cout << "intra ficIndex: " << ficIndex  << " names: '"
             << _drugInteractionList[ficIndex]._names << "' fic: "
             << _drugInteractionList[ficIndex]._fic << endl;
    #endif
    if(ficIndex > 0)
    {
        shouldKill = checkKillInteractAdjDose("Intracellular", ficIndex,
                                              concVec, emaxVec, c50Vec, hVec,
                                              intc50, intEmax, intH,
                                              interactFactor);
    }
    else
    {
        // No drugs met the conditions for drug interaction killing, so use the
        // single drug killing.
        shouldKill = DrugKill::checkKillIntra(mtbAcList, immuneCellAcList, pg, pos, interactFactor);
        interactFactor += 1.0;
    }

    pg.setIntFIC(pos, fic);

    return shouldKill;
}

void DrugKillIndigo::updateStatsIntra(Stats& stats, double mtbLiveCount)
{
    updateStatsIntraNonSingle(stats, mtbLiveCount);
}

void DrugKillIndigo::updateInteractFactorStatIntra(Stats& stats,
                                                   double interactFactor)
{
    updateInteractFactorStatIntraNonSingle(stats, interactFactor);
}

// Get the drug interaction information for intra-cellular drug killing,
// using the dose/adjustment concentration model
//
// An ficIndex into the drug interaction table is defined for the set of drugs
// that meet the killing threshold. If no drugs meet the threshold then it is
// 0.
void DrugKillIndigo::getDrugIntraInteractInfoDoseAdj(
                         const AntibioticConcentrationList& mtbAcList,
                         const AntibioticConcentrationList& immuneCellAcList,
                         vector<double>& adjConc, vector<double>& emaxVec,
                         vector<double>& c50Vec, vector<double>& hVec,
                         double& intEmax, double& intc50, double& intH,
                         int& ficIndex, const GrGrid& pg, const Pos& pos)
{
    double c50 = 0.0;
    double conc = 0.0;
    double effectiveConc = 0.0;
    intEmax = 0.0;
    intc50 = 0.0;
    intH = 0.0;
    ficIndex = 0;

    for (unsigned abxInd=0; abxInd < mtbAcList.size(); abxInd++)
    {
        // We need the mac's concentration object to get the concentration.
        // We need the mtb's concentration object to get the C50 values.
        AntibioticConcentration* pImmuneCellAbxConc = immuneCellAcList.at(abxInd);
        AntibioticConcentration* pMtbAbxConc = mtbAcList.at(abxInd);
        AntibioticSimulationGrid* pAsg = pImmuneCellAbxConc->antibioticSimulationGrid();

        conc = getTotalConc(pMtbAbxConc, pg, pos, abxInd);
        effectiveConc = pMtbAbxConc->effectiveConcentration(conc);

        if(effectiveConc < pMtbAbxConc->getKillThreshold(_PARAM(Drugs_interactThreshold)))
        {
            continue;
        }

        c50 = pMtbAbxConc->mtbSpecificC50Intra();

        // Get all relevant PD information for rate constant
        // calculation
        adjConc.push_back(effectiveConc);
        emaxVec.push_back(pAsg->intraActivity());
        c50Vec.push_back(c50);
        hVec.push_back(pAsg->hillIntra());

        // check to see if the current Emax is larger,
        // if it is, store that as the interaction Emax,
        // and also store that antbiotic's corresponding c50
        if(pAsg->intraActivity() > intEmax){
            intEmax = pAsg->intraActivity();
            intc50 = c50;
            intH = pAsg->hillIntra();
        }

        // Set the bit corresponding to the drug.
        ficIndex += (1 << abxInd);
    }
}

bool DrugKillIndigo::checkKillInteractAdjDose(
                         string locationType, int ficIndex,
                         const vector<double>& concVec,
                         const vector<double>& emaxVec,
                         const vector<double>& c50Vec,
                         const vector<double>& hVec,
                         double intc50, double intEmax, double intH,
                         double& interactFactor)
{
    double fic = _drugInteractionList[ficIndex]._fic;

    if(fic < 0)
    {
        cout << locationType << " fic of " << fic
             << " is < 0, for ficIndex of " << ficIndex << "." << endl
             << "This is for a drug combination without an fic in the drug"
             << " interaction file."
             << endl;
        throw std::runtime_error("Fatal error");
    }
    
    double effectiveConc = AntibioticSimulationGrid::computeInteractAdjConc(fic, concVec, emaxVec, c50Vec, hVec, intc50, intEmax, intH);
    double nonInteractConc = AntibioticSimulationGrid::computeInteractAdjConc(1.0, concVec, emaxVec, c50Vec, hVec, intc50, intEmax, intH);

    double killRate = AntibioticConcentration::killingRateConstFromEmax(
                                                   intEmax, effectiveConc,
                                                   intH, intc50);
    double nonInteractKillRate = AntibioticConcentration::killingRateConstFromEmax(
                                                              intEmax, nonInteractConc,
                                                              intH, intc50);
    interactFactor += safeDiv(killRate, nonInteractKillRate);

    bool shouldKill = AntibioticSimulationGrid::checkShouldKill(killRate);
    return shouldKill;
}
