
#include "drug-kill.h"
#include "grgrid.h"
#include "stat.h"
#include "antibioticconcentration.h"

BOOST_CLASS_EXPORT_IMPLEMENT(DrugKill)

using namespace std;

// Set to 1 for some standard debug output.
#define DBGDK 0

DrugKill::DrugKill()
    : _drugInteractionMethod(DRIM_NONE)
{
    #if DBGDK
    cout << "DrugKill::DrugKill default " << this << endl;
    #endif
}

DrugKill::DrugKill(DrugInteractionMethod method)
    : _drugInteractionMethod(method)
{
    #if DBGDK
    cout << "DrugKill::DrugKill method " << this << endl;
    #endif
}

DrugKill::~DrugKill()
{
    #if DBGDK
    cout << "DrugKill::~DrugKill " << this << endl;
    #endif
}

double DrugKill::getTotalConc(AntibioticConcentration* pCurrentAbxConc,
                              const GrGrid& grid, const Pos& pos, int abxIdx)
{
    AntibioticSimulationGrid* pAbxGrid = pCurrentAbxConc->antibioticSimulationGrid();

    // Get the total concentration, adjusting for a mac if one is present.
    double freeAbx = pAbxGrid->currentGrid().grid(pos);            //  mg/L
    double bndAbx = pAbxGrid->currentGrid().casBnd_grid(pos);      //  mg/L
    double epithBndAbx = pAbxGrid->currentGrid().epithBnd_grid(pos);      //  mg/L
    double conc = freeAbx + bndAbx + epithBndAbx;

    if(grid.hasAgentType(MAC,pos))
    {
        const Mac* pMac = grid.getAgent<Mac>(pos);
        if(!pMac)
        {
            throw runtime_error("Fatal error: DrugKill::getTotalConc, no Mac present in compartment.");
        }
        AntibioticConcentration* pMacAbxConc = pMac->getAntibioticConcentration(abxIdx);

		double intraAbx = pMacAbxConc->concentration();
        double concMg = intraAbx * VOL_MAC + freeAbx * VOL +
                        bndAbx * VOL + epithBndAbx * VOL;
        conc = concMg / (VOL_MAC+VOL);
    }
    return conc;
}

bool DrugKill::checkKillExtra(const AntibioticConcentrationList& abxConcList,
                              MtbGroupLocation location, GrGrid& grid,
                              const Pos& pos, double& /* interactFactor */)
{
    #if DBGDK
    cout << "DrugKill::checkKillExtra start" << endl;
    #endif

    bool shouldKill = false;

    double emax = 0.0;
    double hill = 0.0;
    double c50 = 0.0;
    double conc = 0.0;
    double effectiveConc = 0.0;

    double maxKillRate = 0.0;
    double singleDrugKillRate = 0.0;

    int abxIdx = 0;
    for (AntibioticConcentrationList::const_iterator abxit = abxConcList.begin();
         abxit != abxConcList.end(); ++abxit, ++abxIdx)
    {
        AntibioticConcentration* pCurrentAbxConc = *abxit;

        AntibioticSimulationGrid* asg = pCurrentAbxConc->antibioticSimulationGrid();

        conc = getTotalConc(pCurrentAbxConc, grid, pos, abxIdx);
        effectiveConc = pCurrentAbxConc->effectiveConcentration(conc);

        if (location == NON_REPLICATING_EXTRA_CELLULAR)
        {
            c50 = pCurrentAbxConc->mtbSpecificC50Nonrep();
            emax = pCurrentAbxConc->antibioticSimulationGrid()->nonrepActivity();
            hill = pCurrentAbxConc->antibioticSimulationGrid()->hillNonrep();
        }
        else
        {
            c50 = pCurrentAbxConc->mtbSpecificC50Extra();
            emax = pCurrentAbxConc->antibioticSimulationGrid()->extraActivity();
            hill = pCurrentAbxConc->antibioticSimulationGrid()->hillExtra();
        }

        singleDrugKillRate = AntibioticConcentration::killingRateConstFromEmax(emax,
                                                       effectiveConc, hill, c50);

        // Indifference assumes that the combined effect is the maximum of the
        // individual effects. If only one drug is dosed, this takes the max of
        // the individual killing rates and zero.
        maxKillRate = std::max(singleDrugKillRate, maxKillRate);
    }

    shouldKill = AntibioticSimulationGrid::checkShouldKill(maxKillRate);
    return shouldKill;
}

// Not relevant for single drug killing.
void DrugKill::updateStatsExtra(Stats& /* stats */,
                           MtbGroupLocation /* location */,
                           double /* mtbLiveCount */)
{
}

// Not relevant for single drug killing.
void DrugKill::updateInteractFactorStatExtra(Stats& /* stats */,
                                   MtbGroupLocation /* location */,
                                   double /* interactFactor */)
{
}

// Used by all drug interaction methods other than single drug killing.
void DrugKill::updateStatsExtraNonSingle(Stats& stats, MtbGroupLocation location,
                                      double mtbLiveCount)
{
    // Use the number of live Mtb just before drug killing to keep track
    // of average strength of interaction Mtb experiences
    if(location == NON_REPLICATING_EXTRA_CELLULAR){
        stats.setNumNonPreDrug(stats.getNumNonPreDrug() + mtbLiveCount);
    }
    else{
        stats.setNumExtPreDrug(stats.getNumExtPreDrug() + mtbLiveCount);
    }
}

// Used by all drug interaction methods other than single drug killing.
void DrugKill::updateInteractFactorStatExtraNonSingle(Stats& stats,
                                                      MtbGroupLocation location,
                                                      double interactFactor)
{
    if(location == NON_REPLICATING_EXTRA_CELLULAR){
        stats.setAvgNonInteractFactor(stats.getAvgNonInteractFactor() + interactFactor);
    }
    else{
        stats.setAvgExtInteractFactor(stats.getAvgExtInteractFactor() + interactFactor);
    }
}

bool DrugKill::checkKillIntra(
              const AntibioticConcentrationList& mtbAcList,
              const AntibioticConcentrationList& immuneCellAcList,
              GrGrid& grid, const Pos& pos,
              double& /* interactFactor */)
{
    #if DBGDK
    cout << "DrugKill::checkKillIntra start" << endl;
    #endif

    bool shouldKill = false;

    double emax = 0.0;
    double hill = 0.0;
    double c50 = 0.0;
    double conc = 0.0;
    double effectiveConc = 0.0;

    double maxKillRate = 0.0;
    double singleDrugKillRate = 0.0;
    for (unsigned abxInd=0; abxInd < mtbAcList.size(); abxInd++)
    {
        // We need the immune cell's concentration object to get the concentration.
        // We need the mtb's concentration object to get the C50 values.
        AntibioticConcentration* pImmuneCellAbxConc = immuneCellAcList.at(abxInd);
        AntibioticConcentration* pMtbAbxConc = mtbAcList.at(abxInd);
        AntibioticSimulationGrid* asg = pMtbAbxConc->antibioticSimulationGrid();

        conc = getTotalConc(pMtbAbxConc, grid, pos, abxInd);
        effectiveConc = pMtbAbxConc->effectiveConcentration(conc);

        emax = asg->intraActivity();
        hill = asg->hillIntra();
        c50 = pMtbAbxConc->mtbSpecificC50Intra();

        singleDrugKillRate = AntibioticConcentration::killingRateConstFromEmax(emax,
                                                          effectiveConc, hill, c50);

        // Indifference assumes that the combined effect is the maximum of the
        // individual effects. If only one drug is dosed, this takes the max of
        // the individual killing rates and zero.
        maxKillRate = std::max(singleDrugKillRate, maxKillRate);
    }

    shouldKill = AntibioticSimulationGrid::checkShouldKill(maxKillRate);

    return shouldKill;
}

// Not relevant for single drug killing.
void DrugKill::updateStatsIntra(Stats& /* stats */, double /* mtbLiveCount */)
{
}

// Not relevant for single drug killing.
void DrugKill::updateInteractFactorStatIntra(Stats& /* stats */,
                                             double /* interactFactor */)
{
}

// Used by all drug interaction methods other than single drug killing.
void DrugKill::updateStatsIntraNonSingle(Stats& stats, double mtbLiveCount)
{
    // Use the number of live Mtb just before drug killing to keep track
    // of average strength of interaction Mtb experiences
    stats.setNumIntPreDrug(stats.getNumIntPreDrug() + mtbLiveCount);
}

// Used by all drug interaction methods other than single drug killing.
void DrugKill::updateInteractFactorStatIntraNonSingle(Stats& stats,
                                                      double interactFactor)
{
    stats.setAvgIntInteractFactor(stats.getAvgIntInteractFactor() + interactFactor);
}
