/*
 * GrDiffusionFFT.cpp
 *
 *  Created on: June 7th, 2013
 *      Author: ncilfone
 */

#include <fftw3.h>
#include "grdiffusionFFT.h"
#include "grsimulationgrid.h"
#include "antibioticsimulationgrid.h"


#include <iostream>
#include <cstdlib>
#include <fstream>

//    GENERAL DST/DCT PROCEDURE
//    --------------------------------------------------------------------
//    Algorithm - Zero BC
//    --------------------------------------------------------------------
//    1. Assemble the DST vector (_zeroBCfftVector) from the current grid
//    2. Execute a forward DST using the DST forward plan
//    3. Multiply the forward DsT by pMat for zero BC and a scaling factor
//    4. Execute a reverse DST using the reverse plan
//    5. Disassemble the DST vector to the next grid
//    --------------------------------------------------------------------
//    Algorithm - No Flux
//    --------------------------------------------------------------------
//    1. Assemble the DCT vector (_noFluxfftVector) from the current grid
//    2. Execute a forward DCT using the DCT forward plan
//    3. Multiply the forward DCT by pMat for No Flux and a scaling factor
//    4. Execute a reverse DCT using the DCT reverse plan
//    5. Disassemble the DCT vector to the next grid
//    --------------------------------------------------------------------

using namespace std;

// Set to 1 to diffuse O2.
#define diffuseO2 0

#define PADDED_PMAT(type,name) \
    , _pMat_##name(_dim.size(), 0)

GrDiffusionFFT::GrDiffusionFFT(const Pos dim, DiffusionMethod diffusionMethod)
    : GrDiffusion()
      , _cutOffValue(DIFFUSION_CUTOFF)
      , _dim(dim)
      , _diffusionMethod(diffusionMethod) 
      , _zeroBCfftVector((double*)fftw_malloc(sizeof(double)*_dim.size()))
      , _noFluxfftVector((double*)fftw_malloc(sizeof(double)*_dim.size()))
      PMAT_DEFS
{
	// Define the plans in the constructor so it is only done once.

    std::pair<int, string> planTypeInfo = selectFFTWPlanType(_diffusionMethod);
    int planType = planTypeInfo.first;
    string planTypeName = planTypeInfo.second;

    // Plans for DST
    cout << "Zero Boundary Condition -- Computing the most effective DST algorithm using " << planTypeName << "..." << endl << endl;
	cout << "FFT Prime Grid Size: " << _dim << endl << endl;
    _zeroBCForward = fftw_plan_r2r_2d(_dim.x, _dim.y, _zeroBCfftVector, _zeroBCfftVector, FFTW_RODFT10, FFTW_RODFT10, planType);
    _zeroBCReverse = fftw_plan_r2r_2d(_dim.x, _dim.y, _zeroBCfftVector, _zeroBCfftVector, FFTW_RODFT01, FFTW_RODFT01, planType);

    // Plans for DCT
    cout << "No Flux Boundary Condition -- Computing the most effective DCT algorithm using " << planTypeName << "..." << endl << endl;
	cout << "FFT Prime Grid Size: " << _dim << endl << endl;
    _noFluxForward = fftw_plan_r2r_2d(_dim.x, _dim.y, _noFluxfftVector, _noFluxfftVector, FFTW_REDFT10, FFTW_REDFT10, planType);
    _noFluxReverse = fftw_plan_r2r_2d(_dim.x, _dim.y, _noFluxfftVector, _noFluxfftVector, FFTW_REDFT01, FFTW_REDFT01, planType);

    // Build PMats for molecules that need to be diffused
    calcAllPMat();
}
#undef PADDED_PMAT

#define PADDED_PMAT(type,name) \
    , _pMat_##name(diffFFT._dim.size(), 0)
GrDiffusionFFT::GrDiffusionFFT(const GrDiffusionFFT& diffFFT)
    : GrDiffusion()
      , _cutOffValue(diffFFT._cutOffValue)
      , _dim(diffFFT._dim)
	  , _diffusionMethod(diffFFT._diffusionMethod) 
      , _zeroBCfftVector((double*)fftw_malloc(sizeof(double)*_dim.size()))
      , _noFluxfftVector((double*)fftw_malloc(sizeof(double)*_dim.size()))
      PMAT_DEFS
{
    // Note that this is inefficient, since this copy of GrDiffusionFFT is not
    // likely to be used to compute diffusion.

    // Define the plans in the constructor so it is only done ONCE

    std::pair<int, string> planTypeInfo = selectFFTWPlanType(_diffusionMethod);
    int planType = planTypeInfo.first;
    string planTypeName = planTypeInfo.second;

    // Plans for DST
    //cout << "Zero Boundary Condition -- Computing the most effective DST algorithm using " << planTypeName << "..." << endl << endl;
	//cout << "FFT Prime Grid Size: " << _dim << endl << endl;
    _zeroBCForward = fftw_plan_r2r_2d(_dim.x, _dim.y, _zeroBCfftVector, _zeroBCfftVector, FFTW_RODFT10, FFTW_RODFT10, planType);
    _zeroBCReverse = fftw_plan_r2r_2d(_dim.x, _dim.y, _zeroBCfftVector, _zeroBCfftVector, FFTW_RODFT01, FFTW_RODFT01, planType);

    // Plans for DCT
    //cout << "No Flux Boundary Condition -- Computing the most effective DCT algorithm using " << planTypeName << "..." << endl << endl;
	//cout << "FFT Prime Grid Size: " << _dim << endl << endl;
    _noFluxForward = fftw_plan_r2r_2d(_dim.x, _dim.y, _noFluxfftVector, _noFluxfftVector, FFTW_REDFT10, FFTW_REDFT10, planType);
    _noFluxReverse = fftw_plan_r2r_2d(_dim.x, _dim.y, _noFluxfftVector, _noFluxfftVector, FFTW_REDFT01, FFTW_REDFT01, planType);

    // Build PMats for molecules that need to be diffused
    calcAllPMat();
}
#undef PADDED_PMAT


GrDiffusionFFT::~GrDiffusionFFT()
{
    // Use the FFTw library to free the allocations
    fftw_destroy_plan(_noFluxForward);
    fftw_destroy_plan(_noFluxReverse);
    fftw_destroy_plan(_zeroBCForward);
    fftw_destroy_plan(_zeroBCReverse);
    fftw_free(_noFluxfftVector);
    fftw_free(_zeroBCfftVector);
}

void GrDiffusionFFT::assembleFFTVectorZeroBC(const Scalar* currentGrid, const Pos &dim)
{
    int padIndexMover = 0;
    for(int i = 0; i < (dim.x); i++)  {
        for(int j = 0; j < (dim.y); j++)  {
            padIndexMover = (i*dim.y)+j;
            _zeroBCfftVector[padIndexMover] = currentGrid[GridSpec::padInd(dim, i, j)];
        }
    }
}


void GrDiffusionFFT::assembleFFTVectorNoFlux(const Scalar* currentGrid, const Pos &dim)
{
    int padIndexMover = 0;
    for(int i = 0; i < (dim.x); i++)  {
        for(int j = 0; j < (dim.y); j++)  {
            padIndexMover = (i*dim.y)+j;
            _noFluxfftVector[padIndexMover] = currentGrid[GridSpec::padInd(dim, i, j)];
        }
    }
}

void GrDiffusionFFT::disassembleFFTVectorNoFlux(Scalar* nextGrid, const Pos& dim)
{
    int padIndexMover = 0;
    for(int i = 0; i < (dim.x); i++)  {
        for(int j = 0; j < (dim.y); j++)  {
            padIndexMover = (i*dim.y)+j;
            nextGrid[GridSpec::padInd(dim, i, j)] = _noFluxfftVector[padIndexMover] <= _cutOffValue ? double(0.0) : _noFluxfftVector[padIndexMover];
            // Clean up vector #s
            _noFluxfftVector[padIndexMover] = 0;
        }
    }

}

void GrDiffusionFFT::disassembleFFTVectorZeroBC(Scalar* nextGrid, const Pos& dim)
{
    int padIndexMover = 0;
    for(int i = 0; i < (dim.x); i++)  {
        for(int j = 0; j < (dim.y); j++)  {
            padIndexMover = (i*dim.y)+j;
            nextGrid[GridSpec::padInd(dim, i, j)] = _zeroBCfftVector[padIndexMover] <= _cutOffValue ? double(0.0) : _zeroBCfftVector[padIndexMover];
            // Clean up vector #s
            _zeroBCfftVector[padIndexMover] = 0;
        }
    }
}

void GrDiffusionFFT::degrade(Scalar* nextGrid, Scalar* currentGrid, const Pos& dim, double degRate, double dt)
{
    const Scalar molDegRate = exp(-1.0 * degRate * dt);
    double newVal;
    for(int i = 0; i < (dim.x); i++)  {
        for(int j = 0; j < (dim.y); j++)  {
            newVal = currentGrid[GridSpec::ind(dim, i, j)] * molDegRate;
            nextGrid[GridSpec::ind(dim, i, j)] = newVal <= _cutOffValue ? double(0.0) : newVal;
        }
    }
}

// This function is used to degrade chemicals that are not diffused. No cutoff
// processing is needed.
static void diffuse_degrade(Scalar* __restrict__ newgrid, const Pos& dim, const Scalar degRate)
{
    for(int i = 0; i < dim.x; i++)
    {
      for(int j = 0; j < dim.y; j++)
      {
          Scalar& newconc = newgrid[GridSpec::padInd(dim, i, j)];
          
          const Scalar dt = _PARAM(Diffusion_FFT_timestepDiffusion);
          const Scalar molDegRate = exp(-1.0 * degRate * dt);
          
          newconc *= molDegRate;
      }
    }
}

void GrDiffusionFFT::diffuse(GrSimulationGrid& grSim, const int time, bool tnfrDynamics, bool immunoMod, bool doDrugs)
{
    // Get the current and next grid
    GrGrid& grid = grSim.getCurrentGrid();
    GrGrid& nextGrid = grSim.getNextGrid();

    // Inactive TGFB is not diffused.
	// Copy it to the next grid, so any other changes made during the diffusion
	// loop, such as secretion, are propagated to the next grid.
    nextGrid.TGFBInactiveVector() = grid.TGFBInactiveVector();
    ::diffuse_degrade(nextGrid.TGFBInactive(), grid.getRange(), _PARAM(Core_degRateTGFBInactive));

    bool doImmunoMod = (immunoMod && (time >= _PARAM(Mod_dosageStartTime)));

    // TNF - Zero BC
    assembleFFTVectorZeroBC(grid.TNF(), grid.getRange());
    fftw_execute(_zeroBCForward);
    pMatWithForwardFFTandScalingZeroBC(_pMat_TNF);
    fftw_execute(_zeroBCReverse);
    disassembleFFTVectorZeroBC(nextGrid.TNF(), grid.getRange());

    // Shed TNFR2 - Zero BC
    if(tnfrDynamics)
    {
        assembleFFTVectorZeroBC(grid.shedTNFR2(), grid.getRange());
        fftw_execute(_zeroBCForward);
        pMatWithForwardFFTandScalingZeroBC(_pMat_shedTNFR2);
        fftw_execute(_zeroBCReverse);
        disassembleFFTVectorZeroBC(nextGrid.shedTNFR2(), grid.getRange());
    }

    // CCL2 - Zero BC
    assembleFFTVectorZeroBC(grid.CCL2(), grid.getRange());
    fftw_execute(_zeroBCForward);
    pMatWithForwardFFTandScalingZeroBC(_pMat_CCL2);
    fftw_execute(_zeroBCReverse);
    disassembleFFTVectorZeroBC(nextGrid.CCL2(),grid.getRange());
    diffuse_ratio_CC(nextGrid.CCL2(), nextGrid.CCL5(), nextGrid.CXCL9(), grid.getRange());

    // IL-10 - Zero BC
    assembleFFTVectorZeroBC(grid.il10(), grid.getRange());
    fftw_execute(_zeroBCForward);
    pMatWithForwardFFTandScalingZeroBC(_pMat_il10);
    fftw_execute(_zeroBCReverse);
    disassembleFFTVectorZeroBC(nextGrid.il10(), grid.getRange());

    // TGFBActive - Zero BC
    assembleFFTVectorZeroBC(grid.TGFBActive(), grid.getRange());
    fftw_execute(_zeroBCForward);
    pMatWithForwardFFTandScalingZeroBC(_pMat_TGFBActive);
    fftw_execute(_zeroBCReverse);
    disassembleFFTVectorZeroBC(nextGrid.TGFBActive(), grid.getRange());


    if (doDrugs)  {
        // THESE DO NOT HAVE THE ABILITY TO CHANGE DIFFUSIVITY ON THE FLY YET!!!!

        AntibioticSimulationGridList& antibioticList = grid.antibioticSimulationGridList();

        const Scalar dt = _PARAM(Diffusion_FFT_timestepDiffusion);
        double degRate;
        for (AntibioticSimulationGridList::iterator it = antibioticList.begin(); it != antibioticList.end(); it++)
        {
            if((*it)->dose() == 0.0)
            {
    			// Don't diffuse an antibiotic whose dose is 0. This can can
    			// improve performance. This can happen with drug optimization,
    			// since the optimization algorithm can choose a dose of 0 for
    			// any of the drugs in a drug combination. This happens outside
    			// of the model code.
                continue;
            }

            assembleFFTVectorNoFlux((*it)->currentGrid().grid(), grid.getRange());
            fftw_execute(_noFluxForward);
            pMatWithForwardFFTandScalingNoFlux((*it)->pmat());
            fftw_execute(_noFluxReverse);
            disassembleFFTVectorNoFlux((*it)->nextGrid().grid(), grid.getRange());

            // Caseum and epithelium bound antibiotics do not diffuse, but they do degrade.
            // Copy it to the next grid, so any other changes made during the
            // diffusion loop are propagated to the next grid.
            degRate = (*it)->degRateConst();
            degrade((*it)->nextGrid().casBnd_grid(), (*it)->currentGrid().casBnd_grid(), grid.getRange(), degRate, dt);
            degrade((*it)->nextGrid().epithBnd_grid(), (*it)->currentGrid().epithBnd_grid(), grid.getRange(), degRate, dt);

            (*it)->swapGrids();
        }
    }


    if (doImmunoMod)  {

    // Etanercept - No Flux
    assembleFFTVectorNoFlux(grid.Etanercept(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_Etanercept);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.Etanercept(), grid.getRange());

    // Infliximab - No Flux
    assembleFFTVectorNoFlux(grid.Infliximab(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_Infliximab);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.Infliximab(), grid.getRange());

    // Etanercept:TNF Complex - No Flux
    assembleFFTVectorNoFlux(grid.EtanTNF(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_EtanTNF);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.EtanTNF(), grid.getRange());

    // (Infliximab)1:TNF Complex - No Flux
    assembleFFTVectorNoFlux(grid.Ifx1TNF(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_Ifx1TNF);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.Ifx1TNF(), grid.getRange());

    // (Infliximab)2:TNF Complex - No Flux
    assembleFFTVectorNoFlux(grid.Ifx2TNF(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_Ifx2TNF);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.Ifx2TNF(), grid.getRange());

    // (Infliximab)3:TNF Complex - No Flux
    assembleFFTVectorNoFlux(grid.Ifx3TNF(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_Ifx3TNF);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.Ifx3TNF(), grid.getRange());
}


    #if diffuseO2
    // O2 - No Flux
    assembleFFTVectorNoFlux(grid.O2(), grid.getRange());
    fftw_execute(_noFluxForward);
    pMatWithForwardFFTandScalingNoFlux(_pMat_O2);
    fftw_execute(_noFluxReverse);
    disassembleFFTVectorNoFlux(nextGrid.O2(), grid.getRange());
    #endif


    // Calculate the shedTNFR2 binding and un-binding
    diffuse_shed_bind_unbind(nextGrid, tnfrDynamics);

    grSim.swap();
}

void GrDiffusionFFT::smoother(GrSimulationGrid& grSim, const int time, bool tnfrDynamics, bool immunoMod, bool doDrugs)
{   
    bool doImmunoMod = (immunoMod && (time >= _PARAM(Mod_dosageStartTime)));

    // SMOOTHER STEPS
    for (int i = 0; i < _PARAM(Diffusion_FFT_numSmootherSteps); i++)  {

        // Get the current and next grid
        GrGrid& preSmoothGrid = grSim.getCurrentGrid();
        GrGrid& postSmoothGrid = grSim.getNextGrid();

        // TNF
        forwardEulerSmootherZeroBC(preSmoothGrid.TNF(), postSmoothGrid.TNF(), _PARAM(Core_diffusivityTNF), preSmoothGrid.getRange());

        // Shed TNFR2
        if(tnfrDynamics)
        {
            forwardEulerSmootherZeroBC(preSmoothGrid.shedTNFR2(), postSmoothGrid.shedTNFR2(), _PARAM(Molecular_TNF_diffusivityShedTNFR2), preSmoothGrid.getRange());
        }

        // CCL2
        forwardEulerSmootherZeroBC(preSmoothGrid.CCL2(), postSmoothGrid.CCL2(), _PARAM(Core_diffusivityChemokines), preSmoothGrid.getRange());
        diffuse_ratio_CC(postSmoothGrid.CCL2(), postSmoothGrid.CCL5(), postSmoothGrid.CXCL9(), postSmoothGrid.getRange());

        // IL-10
        forwardEulerSmootherZeroBC(preSmoothGrid.il10(), postSmoothGrid.il10(), _PARAM(Core_diffusivityIL10), preSmoothGrid.getRange());
    
        // TGFB
        forwardEulerSmootherZeroBC(preSmoothGrid.TGFBActive(), postSmoothGrid.TGFBActive(), _PARAM(Core_diffusivityTGFBActive), preSmoothGrid.getRange());


        if (doDrugs)  {
            // CHECK TO MAKE SURE ZERO BC FORWARD EULER IS WORKING PROPERLY!!!!!

            AntibioticSimulationGridList& antibioticList = preSmoothGrid.antibioticSimulationGridList();

            for (AntibioticSimulationGridList::iterator it = antibioticList.begin(); it != antibioticList.end(); it++)
            {
                forwardEulerSmootherNoFlux((*it)->currentGrid().grid(),
                                           (*it)->nextGrid().grid(),
                                           (*it)->diffusivity(),
                                           preSmoothGrid.getRange());
                // The smoother call does not include any calculations that deal with the non diffusing antibiogic grids
                // (caseum bound, and epithelium bound). In order for the next grids to be updated we need to write the current
                // grids to next grids for these non-diffusing grids before we swap grids.
                (*it)->updateNonDiffusingGrids();

                (*it)->swapGrids();
            }

        }
        if (doImmunoMod)  {
            // CHECK TO MAKE SURE ZERO BC FORWARD EULER IS WORKING PROPERLY!!!!!

            // Etanercept
            forwardEulerSmootherNoFlux(preSmoothGrid.Etanercept(), postSmoothGrid.Etanercept(), _PARAM(Mod_Etanercept_diffusivity), preSmoothGrid.getRange());

            // Infliximab
            forwardEulerSmootherNoFlux(preSmoothGrid.Infliximab(), postSmoothGrid.Infliximab(), _PARAM(Mod_Infliximab_diffusivity), preSmoothGrid.getRange());

            // Etanercept:TNF Complex
            forwardEulerSmootherNoFlux(preSmoothGrid.EtanTNF(), postSmoothGrid.EtanTNF(), _PARAM(Mod_EtanTNF_diffusivity), preSmoothGrid.getRange());

            // (Infliximab)1:TNF Complex
            forwardEulerSmootherNoFlux(preSmoothGrid.Ifx1TNF(), postSmoothGrid.Ifx1TNF(), _PARAM(Mod_Ifx1TNF_diffusivity), preSmoothGrid.getRange());

            // (Infliximab)2:TNF Complex
            forwardEulerSmootherNoFlux(preSmoothGrid.Ifx2TNF(), postSmoothGrid.Ifx2TNF(), _PARAM(Mod_Ifx2TNF_diffusivity), preSmoothGrid.getRange());

            // (Infliximab)3:TNF Complex
            forwardEulerSmootherNoFlux(preSmoothGrid.Ifx3TNF(), postSmoothGrid.Ifx3TNF(), _PARAM(Mod_Ifx3TNF_diffusivity), preSmoothGrid.getRange());
        }

        #if diffuseO2
        // O2
        forwardEulerSmootherNoFlux(preSmoothGrid.O2(), postSmoothGrid.O2(), _PARAM(O2_diffusivityO2), preSmoothGrid.getRange());
        #endif

        grSim.swap();
    }
}


void GrDiffusionFFT::forwardEulerSmootherZeroBC(Scalar* currentGrid, Scalar* nextGrid, const Scalar diffuse, const Pos& dim)
{
    register Scalar up, dn, lt, rt, ct;

    // Calculate Fourier Number
    const Scalar fo = (diffuse * _PARAM(Diffusion_FFT_timestepDiffusionSmoother))/(GridSpec::MICROCMPT_FACE_AREA_CENTIMETERS_SQUARED);

    for(int i=0; i < dim.x; i++)  {
        for(int j=0; j < dim.y; j++)  {
            up  = currentGrid[GridSpec::padInd(dim, i-1, j)];
            lt  = currentGrid[GridSpec::padInd(dim, i, j-1)];
            ct  = currentGrid[GridSpec::padInd(dim, i, j)];
            rt  = currentGrid[GridSpec::padInd(dim, i, j+1)];
            dn  = currentGrid[GridSpec::padInd(dim, i+1, j)];

            Scalar& nc = nextGrid[GridSpec::padInd(dim, i, j)];
            nc = ct + (fo * (up + dn + lt + rt - (Scalar(4.0) * ct)));
        }
    }
}


void GrDiffusionFFT::forwardEulerSmootherNoFlux(Scalar* currentGrid, Scalar* nextGrid, const Scalar diffuse, const Pos &dim)
{
    register Scalar up, dn, lt, rt, ct;

    // Calculate Fourier Number
    const Scalar fo = (diffuse * _PARAM(Diffusion_FFT_timestepDiffusionSmoother))/(GridSpec::MICROCMPT_FACE_AREA_CENTIMETERS_SQUARED);

    for(int i=0; i < dim.x; i++)  {
        for(int j=0; j < dim.y; j++)  {

            ct  = currentGrid[GridSpec::padInd(dim, i, j)];
            up  = i-1 < 0 ? ct : currentGrid[GridSpec::padInd(dim, i-1, j)];
            lt  = j-1 < 0 ? ct : currentGrid[GridSpec::padInd(dim, i, j-1)];
            rt  = j+1 > dim.y-1 ? ct : currentGrid[GridSpec::padInd(dim, i, j+1)];
            dn  = i+1 > dim.x-1 ? ct : currentGrid[GridSpec::padInd(dim, i+1, j)];

            Scalar& nc = nextGrid[GridSpec::padInd(dim, i, j)];
            nc = ct + (fo * (up + dn + lt + rt - (Scalar(4.0) * ct)));
        }
    }
}

void GrDiffusionFFT::diffuse_shed_bind_unbind(GrGrid& nextGrid, bool tnfrDynamics)
{
    Pos p;
    const Scalar dt = _PARAM(Diffusion_FFT_timestepDiffusion);

    for(p.x = 0; p.x < nextGrid.getRange().x; ++p.x)  {
        for(p.y = 0; p.y < nextGrid.getRange().y; ++p.y)  {

			if (_PARAM(Mod_Etanercept_dosage)) {
                //Etanercept Binding/Unbinding
                double netEtanTNFBinding = ((_PARAM(Mod_Etanercept_kOn) * nextGrid.Etanercept(p) * nextGrid.TNF(p))/NAV/VOL - ( _PARAM(Mod_Etanercept_kOff) * nextGrid.EtanTNF(p))) * dt;
    
                // Etanercept:sTNF Degradation = sTNF only (drug remains)
                double stnfdeg = (_PARAM(Core_kDeg) * nextGrid.EtanTNF(p)) * dt;
                nextGrid.incEtanTNF(p, (netEtanTNFBinding - stnfdeg));
                nextGrid.incEtanercept(p, (stnfdeg - netEtanTNFBinding));
                nextGrid.incTNF(p, (netEtanTNFBinding));
            }

            if (_PARAM(Mod_Infliximab_dosage)) {
                // Infliximab Binding/Unbinding
                double netIfx1TNFBinding = ((_PARAM(Mod_Infliximab_kOn) * nextGrid.Infliximab(p) * nextGrid.TNF(p))/NAV/VOL - ( _PARAM(Mod_Infliximab_kOff) * nextGrid.Ifx1TNF(p))) * dt;
                double netIfx2TNFBinding = ((_PARAM(Mod_Infliximab_kOn) * nextGrid.Infliximab(p) * nextGrid.Ifx1TNF(p))/NAV/VOL - ( _PARAM(Mod_Infliximab_kOff) * nextGrid.Ifx2TNF(p))) * dt;
                double netIfx3TNFBinding = ((_PARAM(Mod_Infliximab_kOn) * nextGrid.Infliximab(p) * nextGrid.Ifx2TNF(p))/NAV/VOL - ( _PARAM(Mod_Infliximab_kOff) * nextGrid.Ifx3TNF(p))) * dt;

                // Infliximab:sTNF Degradation = sTNF only (drug remains)
                double stnf1deg = (_PARAM(Core_kDeg) * nextGrid.Ifx1TNF(p)) * dt;
                double stnf2deg = (_PARAM(Core_kDeg) * nextGrid.Ifx2TNF(p)) * dt;
                double stnf3deg = (_PARAM(Core_kDeg) * nextGrid.Ifx3TNF(p)) * dt;

                nextGrid.incIfx1TNF(p, (netIfx1TNFBinding - stnf1deg));
                nextGrid.incIfx2TNF(p, (netIfx2TNFBinding - stnf2deg));
                nextGrid.incIfx3TNF(p, (netIfx3TNFBinding - stnf3deg));
                nextGrid.incInfliximab(p, (stnf1deg + 2*stnf2deg + 3*stnf3deg - netIfx1TNFBinding - netIfx2TNFBinding - netIfx3TNFBinding));
                nextGrid.incTNF(p, (-netIfx1TNFBinding));

            }

            if(tnfrDynamics)
            {
                // Unbinding of sTNF from ShedTNFR2
                nextGrid.incshedTNFR2(p, (-1.0 * nextGrid.shedTNFR2(p) * _PARAM(Molecular_TNF_KD2) * _PARAM(Molecular_TNF_kOn2) * dt));
    
                // Addition of sTNF from unbinding from ShedTNFR2
                nextGrid.incTNF(p, (1.0 * nextGrid.shedTNFR2(p) * _PARAM(Molecular_TNF_KD2) * _PARAM(Molecular_TNF_kOn2) * dt));
            }
        }
    }
}

void GrDiffusionFFT::pMatWithForwardFFTandScalingNoFlux(const vector<double>& pMat)
{
    // Scaling factor is needed since FFTw
    // calculates un-normalized DCTs
    int size = _dim.size();
    double scaling = 1.0/(4*(double(size)));

    for (int i = 0; i < size; i++)  {
        _noFluxfftVector[i]= _noFluxfftVector[i] * pMat[i] * scaling;
    }
}

void GrDiffusionFFT::pMatWithForwardFFTandScalingZeroBC(const vector<double>& pMat)
{
    // Scaling factor is needed since FFTw
    // calculates un-normalized DCTs
    int size = _dim.size();
    double scaling = 1.0/(4*(double(size)));

    for (int i = 0; i < size; i++)  {
        _zeroBCfftVector[i]= _zeroBCfftVector[i] * pMat[i] * scaling;
    }
}

void GrDiffusionFFT::calcAllPMat()
{
    // Set-up all pMat -- Call this after diffusion
    // object is created and FFT is chosen
    // NOTE: Degradation is built into PMat

    // -- Zero BC PMat
    setupPMATZeroBC(_pMat_TNF, _PARAM(Core_diffusivityTNF), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Core_kDeg));
    setupPMATZeroBC(_pMat_shedTNFR2, _PARAM(Molecular_TNF_diffusivityShedTNFR2), _PARAM(Diffusion_FFT_timestepDiffusion), 0.0); // shedTNFR2 doesn't degrade due to unbinding
    setupPMATZeroBC(_pMat_CCL2, _PARAM(Core_diffusivityChemokines), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Core_ChemokinekDeg));
    setupPMATZeroBC(_pMat_il10, _PARAM(Core_diffusivityIL10), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Core_Ikdeg));
    setupPMATZeroBC(_pMat_TGFBActive,   _PARAM(Core_diffusivityTGFBActive), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Core_degRateTGFBActive));
    setupPMATZeroBC(_pMat_TGFBInactive, _PARAM(Core_diffusivityTGFBActive), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Core_degRateTGFBInactive));
// -- No Flux PMat - Antibiotics
    // Pmats for antibiotics are now being calculated as part of the antibiotic constructor in antibioticsimulationgrid.cpp
    
    // -- No Flux PMat - Other molecules
    #if diffuseO2
    setupPMATNoFlux(_pMat_O2, _PARAM(O2_diffusivityO2), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(O2_degRateConstO2));
    #endif

    setupPMATNoFlux(_pMat_Etanercept, _PARAM(Mod_Etanercept_diffusivity), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Mod_Etanercept_degRateConst));
    setupPMATNoFlux(_pMat_Infliximab, _PARAM(Mod_Infliximab_diffusivity), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Mod_Infliximab_degRateConst));
    setupPMATNoFlux(_pMat_EtanTNF, _PARAM(Mod_EtanTNF_diffusivity), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Mod_EtanTNF_degRateConst));
    setupPMATNoFlux(_pMat_Ifx1TNF, _PARAM(Mod_Ifx1TNF_diffusivity), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Mod_Ifx1TNF_degRateConst));
    setupPMATNoFlux(_pMat_Ifx2TNF, _PARAM(Mod_Ifx2TNF_diffusivity), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Mod_Ifx2TNF_degRateConst));
    setupPMATNoFlux(_pMat_Ifx3TNF, _PARAM(Mod_Ifx3TNF_diffusivity), _PARAM(Diffusion_FFT_timestepDiffusion), _PARAM(Mod_Ifx3TNF_degRateConst));
}

void GrDiffusionFFT::setupPMATZeroBC(vector<double>& pMat, const double diff, const double timestep, const double kdeg)
{
    double PTemp = 0;
    double gridSpace = GridSpec::MICROCMPT_LENGTH_CENTIMETERS;
    int rowMajor = 0;
    for (int i = 0; i < _dim.x; i++)  {
        for (int j = 0; j < _dim.y; j++)  {
            PTemp = ((2.0 * diff)/(gridSpace * gridSpace)) * ( 2.0 + ((kdeg*gridSpace*gridSpace)/(2.0*diff)) - ( cos( ((PI*(j+1)) / (_dim.y+1)) ) + cos( ((PI*(i+1)) / (_dim.x+1)) ) ) );
//            PTemp = ((2.0 * diff)/(gridSpace * gridSpace)) * ( 2.0 - ( cos( ((PI*(j+1)) / (_dim.y+1)) ) + cos( ((PI*(i+1)) / (_dim.x+1)) ) ) );
            rowMajor = (i*(_dim.y) + j);
            pMat[rowMajor] = exp(-PTemp*timestep);
        }
    }
}

void GrDiffusionFFT::setupPMATNoFlux(vector<double>& pMat, const double diff, const double timestep, const double kdeg)
{
    double PTemp = 0;
    double gridSpace = GridSpec::MICROCMPT_LENGTH_CENTIMETERS;
    int rowMajor = 0;
    for (int i = 0; i < _dim.x; i++)  {
        for (int j = 0; j < _dim.y; j++)  {
            PTemp = ((2.0 * diff)/(gridSpace * gridSpace)) * ( 2.0 + ((kdeg*gridSpace*gridSpace)/(2.0*diff)) - ( cos( ((PI*j) / (_dim.y-1)) ) + cos( ((PI*i) / (_dim.x-1)) ) ) );
//            PTemp = ((2.0 * diff)/(gridSpace * gridSpace)) * ( 2.0 - ( cos( ((PI*j) / (_dim.y-1)) ) + cos( ((PI*i) / (_dim.x-1)) ) ) );
            rowMajor = (i*(_dim.x) + j);
            pMat[rowMajor] = exp(-PTemp*timestep);
        }
    }
}

/* DBG - For Writing to CSV While Debugging
    // Define Array of Correct Length For Outputs - FFT
    double gridOutReal[dim.x][dim.y];

    // Grab the output of the DFT
    // and turn it back into array format
    for (int xlength = 0; xlength < dim.x; xlength++)  {
        for (int ylength = 0; ylength < dim.y; ylength++)  {
            int stride = (xlength*dim.y)+ylength;
            gridOutReal[xlength][ylength] = _noFluxfftVector[stride];
        }
    }

    // Write all the outputs to CSV
    ofstream outputfirst;
    outputfirst.open("FFTOut-Assemble.csv");

    for (int writeX = 0; writeX < dim.x; writeX++)  {
        for (int writeY = 0; writeY < dim.y; writeY++)  {
            outputfirst << gridOutReal[writeX][writeY] << ",";
        }
        outputfirst << "\n";
    }
    outputfirst.close();
*/
