/*
 * grvascular.cpp
 *
 * Created by ncilfone
 * 09/25/2012
 */


#include "grvascular.h"
#include "grsimulation.h"
#include "gr.h"

using namespace std;

Vascular::Vascular()
    :
      _bloodConcentrationIL10(0.0), // blood concentration of IL10 (ug/kg)
      _bloodConcentrationTNF(0.0), // blood concentration of TNF (ug/kg)
      _bloodConcentrationEtan(0.0), // blood concentration of Etanercept (ug/kg)
      _bloodConcentrationInfliximab(0.0), // blood concentration of Infliximab (ug/kg)
      _bloodConcentrationEtanTNF(0.0), // blood concentration of Etanercept:TNF complex (ug/kg)
      _bloodConcentrationIfx1TNF(0.0), // blood concentration of (Infliximab)1:TNF complex (ug/kg)
      _bloodConcentrationIfx2TNF(0.0), // blood concentration of (Infliximab)2:TNF complex (ug/kg)
      _bloodConcentrationIfx3TNF(0.0), // blood concentration of (Infliximab)3:TNF complex (ug/kg)
      _cutOffValue(DIFFUSION_CUTOFF)
{
}

Vascular::~Vascular()
{
}

Vascular* Vascular::clone()
{
    Vascular* pVascular = new Vascular(*this);
    return pVascular;
}

void Vascular::solveVascularSources(GrSimulation &grSim, GrGrid& grid,
                                    double dt, const int time, int DiffStep,
                                    const double timestepDiffusion,
                                    ODESolvers::ODEMethod odeSolver)
{
    int realTime = static_cast<int>(time * AGENT_TIME_STEP + DiffStep * timestepDiffusion);

    if (isDoseStartTime(realTime))
    {
        AntibioticSimulationGridList& antibioticList = grid.antibioticSimulationGridList();

        double dose;
        int interval;
        int dosageStartTime = _PARAM(Drugs_dosageStartTime);
        for (AntibioticSimulationGridList::iterator it = antibioticList.begin(); it != antibioticList.end(); it++)
        {
            dose = (*it)->dose();
            interval = (*it)->dosageInterval();
            double& absComp1Concentration = (*it)->getabsComp1Concentration();
            addDose(realTime, dose, interval, absComp1Concentration, dosageStartTime);
        }

        const std::vector<Pos>& sources = grid.getSources();

        double localConc;
        double MW;
        double change;
        double bloodConc;
        double perm;
        double permCoeff;
        double vesselActivity;

        for (PosVector::const_iterator it = sources.begin(); it != sources.end(); it++)
        {
            // From images that J Matilla sent there are no visible vascular
            // sources in the macrophage region or caseum so exclude vessels in
            // high cell density areas and caseum from adding drugs to the
            // grid.
            if (grid.getCellDensity(*it) < 0.8 && !grid.isCaseated(*it))  {

                for (AntibioticSimulationGridList::iterator abxit = antibioticList.begin(); abxit != antibioticList.end(); abxit++)
                {
		            double gridConcBefore = (*abxit)->currentGrid().grid(*it); //DBG drug dosed into caseation?

                    MW = (*abxit)->MW();
                    bloodConc = (*abxit)->getBloodConcentration();
                    perm = (*abxit)->vascularPermeability();
                    permCoeff = (*abxit)->permCoeff();
                    localConc = getAverageOfNeighbours((*abxit)->currentGrid().grid(), *it, grid.getRange(), MW);
                    change = calculateFluxChangeDrugs(localConc, bloodConc, dt, perm, MW, permCoeff, *it);
                    change = std::max(change, -(*abxit)->currentGrid().grid(*it));    // to prevent getting negative values
                    (*abxit)->currentGrid().incGrid(*it, change);

                    if((*abxit)->currentGrid().grid(*it) <= _cutOffValue)
                    {
                        (*abxit)->currentGrid().setGrid(*it, 0.0);
                    }

//					std::cout<<"DBG Flux - Location of the source: "<< *it <<std::endl;
//					std::cout<<"DBG Flux - Free concentration before flux: "<< gridConcBefore <<std::endl;
//					std::cout<<"DBG Flux - Flux: "<< change <<std::endl;
//					std::cout<<"DBG Flux - Free concentration after flux: "<< (*abxit)->currentGrid().grid(*it) <<std::endl;
                }
            }
            else if (grid.getCellDensity(*it) >= 0.8 && !grid.isCaseated(*it))  {
                for (AntibioticSimulationGridList::iterator abxit = antibioticList.begin(); abxit != antibioticList.end(); abxit++)
                {
		            double gridConcBefore = (*abxit)->currentGrid().grid(*it); //DBG drug dosed into caseation?

                    MW = (*abxit)->MW();
                    bloodConc = (*abxit)->getBloodConcentration();
                    perm = (*abxit)->vascularPermeability();
                    permCoeff = (*abxit)->permCoeff();
                    vesselActivity = (*abxit)->vesselActivity();
                    localConc = getAverageOfNeighbours((*abxit)->currentGrid().grid(), *it, grid.getRange(), MW);
                    change = calculateFluxChangeDrugs(localConc, bloodConc, dt, perm, MW, permCoeff, *it);

                    change = change * vesselActivity;         // allows you to tune the amount of drugs coming in and out of vessels in dense cellular areas.
                    change = std::max(change, -(*abxit)->currentGrid().grid(*it));    // to prevent getting negative values

                    if((*abxit)->currentGrid().grid(*it) <= _cutOffValue)
                    {
                        (*abxit)->currentGrid().setGrid(*it, 0.0);
                    }

                    (*abxit)->currentGrid().incGrid(*it, change);
//					std::cout<<"DBG Flux - Location of the source: "<< *it <<std::endl;
//					std::cout<<"DBG Flux - Free concentration before flux: "<< gridConcBefore <<std::endl;
//					std::cout<<"DBG Flux - Flux: "<< change <<std::endl;
//					std::cout<<"DBG Flux - Free concentration after flux: "<< (*abxit)->currentGrid().grid(*it) <<std::endl;
                }
            }
        }

        for (AntibioticSimulationGridList::iterator it = antibioticList.begin(); it != antibioticList.end(); it++)
        {
            double& rbloodConc = (*it)->getBloodConcentration();
            double& absComp1Concentration = (*it)->getabsComp1Concentration();
            double& absComp2Concentration = (*it)->getabsComp2Concentration();
            double& perCompConcentration = (*it)->getPeriphConcentration();
            double kAbs = (*it)->kAbs();
            double q = (*it)->q();
            double volDistCentral = (*it)->volDistCentral();
            double volDistPeriph = (*it)->volDistPeriph();
            double clearance = (*it)->clearance();
            MW = (*it)->MW();
            if ((*it)->PKModel() == PKM_2_COMPARTMENT_2_ABSORPTION)
            {
                std::valarray<double>&  initVector = (*it)->getInitVector();
                updateBloodODE(rbloodConc, absComp1Concentration, absComp2Concentration, perCompConcentration,
                               kAbs, q, volDistCentral, volDistPeriph, clearance,
                               MW, dt, odeSolver, initVector);
            }
            else if ((*it)->PKModel() == PKM_2_COMPARTMENT_1_ABSORPTION)
            {
                std::valarray<double>&  initVector = (*it)->getInitVector();
                updateBloodODE(rbloodConc, absComp1Concentration, perCompConcentration,
                               kAbs, q, volDistCentral, volDistPeriph, clearance,
                               MW, dt, odeSolver, initVector);

            }
    		else
    		{
     		    stringstream ss;
      		    ss << "Invalid PK model of " << (*it)->PKModel() << " for drug " << (*it)->getName();
     		    throw std::runtime_error(ss.str());
    		}
        }

     }
}

void Vascular::updateState()
{
}

void Vascular::solveVascularSourcesMod(GrGrid& grid, double dt,
                                       const int time, int DiffStep,
                                       const double timestepDiffusion)
{
    int realTime = static_cast<int>(time * AGENT_TIME_STEP + DiffStep * timestepDiffusion);

    if (isDoseStartTimeMod(realTime))  {
        addDose(realTime, _PARAM(Mod_TNF_dosage), _PARAM(Mod_TNF_dosageInterval), _bloodConcentrationTNF, _PARAM(Mod_dosageStartTime));
        addDose(realTime, _PARAM(Mod_IL10_dosage), _PARAM(Mod_IL10_dosageInterval), _bloodConcentrationIL10, _PARAM(Mod_dosageStartTime));
        addDose(realTime, _PARAM(Mod_Etanercept_dosage), _PARAM(Mod_Etanercept_dosageInterval), _bloodConcentrationEtan, _PARAM(Mod_dosageStartTime));
        //_bloodConcentrationEtan = _PARAM(Mod_Etanercept_bloodConc);
        addDose(realTime, _PARAM(Mod_Infliximab_dosage), _PARAM(Mod_Infliximab_dosageInterval), _bloodConcentrationInfliximab, _PARAM(Mod_dosageStartTime));

        const std::vector<Pos>& sources = grid.getSources();

        for (PosVector::const_iterator it = sources.begin(); it != sources.end(); it++)  {

            if (grid.getCellDensity(*it) < 0.8 && !grid.isCaseated(*it))  {

                // TNF
                if (_PARAM(Mod_TNF_dosage)) {
                double localTNF = getAverageOfNeighbours(grid.TNF(), *it, grid.getRange(), NAV); // In Moles
                double TNFChange = calculateFluxChange(localTNF, _bloodConcentrationTNF, dt, _PARAM(Mod_TNF_vascularPermeability), MW_TNF, _PARAM(Mod_TNF_volumeDist)); // in mol/gridspace
                grid.incTNF(*it, (TNFChange * NAV)); // # of molecules
                //modulateBlood(_bloodConcentrationTNF, TNFChange, MW_TNF, _PARAM(Mod_TNF_volumeDist));

                }

                // IL-10
                if (_PARAM(Mod_IL10_dosage)) {
                double localIL10 = getAverageOfNeighbours(grid.il10(), *it, grid.getRange(), NAV); // In Moles
                double IL10Change = calculateFluxChange(localIL10, _bloodConcentrationIL10, dt, _PARAM(Mod_IL10_vascularPermeability), MW_IL10, _PARAM(Mod_IL10_volumeDist)); // in mol/gridspace
                grid.incil10(*it, (IL10Change * NAV)); // # of molecules
                //modulateBlood(_bloodConcentrationIL10, IL10Change, MW_IL10, _PARAM(Mod_IL10_volumeDist));

                }

                // Etanercept
                if (_PARAM(Mod_Etanercept_dosage)) {
                double localEtanercept = getAverageOfNeighbours(grid.Etanercept(), *it, grid.getRange(), NAV); // In Moles
                double EtanerceptChange = calculateFluxChange(localEtanercept, _bloodConcentrationEtan, dt, _PARAM(Mod_Etanercept_vascularPermeability), MW_Etanercept, _PARAM(Mod_Etanercept_volumeDist)); // in mol/gridspace
                grid.incEtanercept(*it, (EtanerceptChange * NAV)); // # of molecules
                //modulateBlood(_bloodConcentrationEtan, EtanerceptChange, MW_Etanercept, _PARAM(Mod_Etanercept_volumeDist));

                // DBG DOSE
                // cerr << "at time" << time << "Blood Concentration" << _bloodConcentrationEtan << endl;
                }

                // Infliximab
                if (_PARAM(Mod_Infliximab_dosage)) {
                double localInfliximab = getAverageOfNeighbours(grid.Infliximab(), *it, grid.getRange(), NAV); // In Moles
                double InfliximabChange = calculateFluxChange(localInfliximab, _bloodConcentrationInfliximab, dt, _PARAM(Mod_Infliximab_vascularPermeability), MW_Infliximab, _PARAM(Mod_Infliximab_volumeDist)); // in mol/gridspace
                grid.incInfliximab(*it, (InfliximabChange * NAV)); // # of molecules
                //modulateBlood(_bloodConcentrationInfliximab, InfliximabChange, MW_Infliximab, _PARAM(Mod_Infliximab_volumeDist));
                }
            }
        }

        updateBlood1Compartment(_bloodConcentrationTNF, _PARAM(Mod_TNF_clearanceRate), dt);
        updateBlood1Compartment(_bloodConcentrationIL10, _PARAM(Mod_IL10_clearanceRate), dt);
        updateBlood1Compartment(_bloodConcentrationEtan, _PARAM(Mod_Etanercept_clearanceRate), dt);
        updateBlood1Compartment(_bloodConcentrationInfliximab, _PARAM(Mod_Infliximab_clearanceRate), dt);

    }

}

void Vascular::solveVascularSourcesO2(GrGrid& grid, double dt)
{
    const std::vector<Pos>& sources = grid.getSources();

    for (PosVector::const_iterator it = sources.begin(); it != sources.end(); it++)
    {
        // From images that J Matilla sent there are no visible vascular sources in the macrophage
        //      region or caseum so exclude vessels in high cell density areas and caseum from adding O2
        //      to the grid.
        if (grid.getCellDensity(*it) < 0.8 && !grid.isCaseated(*it))  {
            double localO2 = grid.O2(*it)/MW_O2;   // mol/L
            double O2Change = calculateFluxChangeDrugs(localO2, _PARAM(O2_bloodConcentration), dt, _PARAM(O2_vascularPermeabilityO2), MW_O2, _PARAM(O2_permCoeffO2), *it);
            grid.incO2(*it, O2Change);
        }
    }
}

double Vascular::getAverageOfNeighbours(Scalar* grid, const Pos& pos, const Pos& dim, const Scalar MW)
{
    double sumC = 0.0;
	GridSpec::VonNeumannRegularIterator vnri = GridSpec::vnriBegin(dim, pos);
	GridSpec::VonNeumannRegularIterator vnriEnd = GridSpec::vnriEnd(dim, pos);

    int numNeighbors = 0;
	for(; vnri != vnriEnd; ++vnri)
	{
	  const Pos& p = *vnri;
      sumC += grid[GridSpec::padInd(dim, p)];
      ++numNeighbors;
    }

	// Divide by molecular weight to get moles/L.
	sumC = sumC/numNeighbors/MW;

    return sumC; // Returns average concentration mol/L
}

void Vascular::updateBlood1Compartment(double& bloodconc, const double clearance, const double dt)
{
//    const Scalar degRate = exp(-1.0 *  clearance * dt);
//    bloodconc = bloodconc * degRate;

    bloodconc = bloodconc * (1 - clearance * dt);
}

// Blood ODEs for oral dosing of INH and RIF
void Vascular::updateBloodODE(double& bloodConcentration, double& absComp1Concentration, double& absComp2Concentration, double& perCompConcentration,
                              double ka, double Q, double VDc, double VDp, double CL,
                              double MW, double dt, ODESolvers::ODEMethod odeSolver, std::valarray<double>& initvector)
{

    //_____________________________________________________________________________________________________
    // Drug dynamics in plasma is based on a two-compartment plasma PK model, with two-compartment
    // transit absorption included (as used by Kjellsson et al., 2012).
    // The compartments are referred to as follows:
    // absComp1 and absComp2: 1st and 2nd absorption compartments
    // perComp              : peripheral distribution compartment
    // units of concentration in these compartments are in mg/kg to allow ease of use with PK parameters from other drugs.
    // although the bloodConcentration is in mol/L, it will be converted to mg/kg in the ODE
    // bloodConcentration   : concentration in blood in mol/L and used for updating the grid
    // we do not include their lung and lesion compartments since that is what our grid will be calculating
    //_____________________________________________________________________________________________________

    double blComp = bloodConcentration * MW * VDc;      // units mg/kg (mol/L * mg/mol * L/kg)

    writeValarrayFromMembers(initvector, blComp, absComp1Concentration, absComp2Concentration, perCompConcentration);
    ODESolvers::Stepper* stepper = getStepper(odeSolver, initvector.size());
    static BloodFunc::Params_t params;
    params.ka = &ka;
    params.Q = &Q;
    params.VDc = &VDc;
    params.VDp = &VDp;
    params.CL = &CL;
    BloodFunc& fn = *getDerivFunc();

    stepper->step(initvector, fn, 0.0, dt/3600, (void*)&params);
    writeMembersFromValarray(initvector, MW, VDc, bloodConcentration,
                             absComp1Concentration, absComp2Concentration, perCompConcentration);

}

// Blood ODEs for oral dosing of fluoroquinolones MXF, LVX, GFX
void Vascular::updateBloodODE(double& bloodConcentration, double& absComp1Concentration, double& perCompConcentration,
                              double ka, double Q, double VDc, double VDp, double CL,
                              double MW, double dt, ODESolvers::ODEMethod odeSolver, std::valarray<double>& initvector)
{

    //_____________________________________________________________________________________________________
    // Drug dynamics in plasma is based on a two-compartment plasma PK model with one absorption compartment.
    // The absorption compartment is only to put the dose into. This model is used by fluoroquinolones and is
    // DIFFERENT from the model used for INH and RIF (as used by Kjellsson et al., 2012).
    // The compartments are referred to as follows:
    // perComp              : peripheral distribution compartment
    // units of concentration in these compartments are in mg/kg to allow ease of use with PK parameters from other drugs.
    // although the bloodConcentration is in mol/L, it will be converted to mg/kg in the ODE
    // bloodConcentration   : concentration in blood in mol/L and used for updating the grid
    //_____________________________________________________________________________________________________

    double blComp = bloodConcentration * MW * VDc;      // units mg/kg (mol/L * mg/mol * L/kg)

    writeValarrayFromMembers(initvector, blComp, absComp1Concentration, perCompConcentration);
    ODESolvers::Stepper* stepper = getStepper(odeSolver, initvector.size());
    static BloodFuncTwoComp::Params_t params;
    params.ka = &ka;
    params.Q = &Q;
    params.VDc = &VDc;
    params.VDp = &VDp;
    params.CL = &CL;
    BloodFuncTwoComp& fn = *getDerivFuncTwoComp();//function asks for command line option bools, but they haven't been passed in this far yet!!
    //DBGAGO
    stepper -> step(initvector, fn, 0.0, dt/3600, (void*)&params );
    writeMembersFromValarray(initvector, MW, VDc, bloodConcentration,
                             absComp1Concentration, perCompConcentration);

}


void Vascular::writeValarrayFromMembers(valarray<double>& inputVector, double blComp,
                                        double& absComp1Concentration, double& absComp2Concentration, double& perCompConcentration)
{
    // Defining the initial vector from the vascular private variables

    inputVector[0] = absComp1Concentration;
    inputVector[1] = absComp2Concentration;
    inputVector[2] = perCompConcentration;
    inputVector[3] = blComp;
}

void Vascular::writeValarrayFromMembers(valarray<double>& inputVector, double blComp,
                                        double& absComp1Concentration, double& absComp2Concentration, double& perCompConcentration, double& lungCompConcentration)
{
    // Defining the initial vector from the vascular private variables

    inputVector[0] = absComp1Concentration;
    inputVector[1] = absComp2Concentration;
    inputVector[2] = perCompConcentration;
    inputVector[3] = blComp;
    inputVector[4] = lungCompConcentration;
}

void Vascular::writeValarrayFromMembers(valarray<double>& inputVector, double blComp,
                                        double& absComp1Concentration, double& perCompConcentration)
{
    // Defining the initial vector from the vascular private variables

    inputVector[0] = absComp1Concentration;
    inputVector[1] = perCompConcentration;
    inputVector[2] = blComp;
}

void Vascular::writeMembersFromValarray(valarray<double>& inputVector, double MW, double VDc, double& bloodConcentration,
                                        double& absComp1Concentration, double& absComp2Concentration, double& perCompConcentration)
{
    absComp1Concentration = applyCutoff(inputVector[0], _cutOffValue);
    absComp2Concentration = applyCutoff(inputVector[1], _cutOffValue);
    perCompConcentration = applyCutoff(inputVector[2], _cutOffValue);
    double blComp = inputVector[3];
    bloodConcentration = blComp / MW / VDc;
    bloodConcentration = applyCutoff(bloodConcentration, _cutOffValue);
}

void Vascular::writeMembersFromValarray(valarray<double>& inputVector, double MW, double VDc, double& bloodConcentration,
                                        double& absComp1Concentration, double& absComp2Concentration, double& perCompConcentration, double& lungCompConcentration)
{
    absComp1Concentration = applyCutoff(inputVector[0], _cutOffValue);
    absComp2Concentration = applyCutoff(inputVector[1], _cutOffValue);
    perCompConcentration = applyCutoff(inputVector[2], _cutOffValue);
    double blComp = inputVector[3];
    bloodConcentration = blComp / MW / VDc;
    bloodConcentration = applyCutoff(bloodConcentration, _cutOffValue);
    lungCompConcentration = applyCutoff(inputVector[4], _cutOffValue);
}

void Vascular::writeMembersFromValarray(valarray<double>& inputVector, double MW, double VDc, double& bloodConcentration,
                                        double& absComp1Concentration, double& perCompConcentration)
{
    absComp1Concentration = applyCutoff(inputVector[0], _cutOffValue);
    perCompConcentration = applyCutoff(inputVector[1], _cutOffValue);
    double blComp = inputVector[2];
    bloodConcentration = blComp / MW / VDc;
    bloodConcentration = applyCutoff(bloodConcentration, _cutOffValue);
}

void BloodFunc::operator()(const ODESolvers::ODEState& vecread, double, ODESolvers::Derivative& vecwrite, void* params) const
{
    const double& ka = *(((Params_t*)params)->ka);
    const double& Q = *(((Params_t*)params)->Q);
    const double& VDc = *(((Params_t*)params)->VDc);
    const double& VDp = *(((Params_t*)params)->VDp);
    const double& CL = *(((Params_t*)params)->CL);

//    0 = absComp1
//    1 = absComp2
//    2 = perComp
//    3 = bloodComp

    vecwrite[0] = -ka * vecread[0];
    vecwrite[1] = ka * (vecread[0] - vecread[1]);
    vecwrite[2] = Q * ( vecread[3]/VDc - vecread[2]/VDp);
    vecwrite[3] = ka * vecread[1] - Q * ( vecread[3]/VDc - vecread[2]/VDp) - CL*vecread[3]/VDc;
}

void BloodFuncLung::operator()(const ODESolvers::ODEState& vecread, double, ODESolvers::Derivative& vecwrite, void* params) const
{
    const double& ka = *(((Params_t*)params)->ka);
    const double& Qp = *(((Params_t*)params)->Qp);
    const double& Qn = *(((Params_t*)params)->Qn);
    const double& VDc = *(((Params_t*)params)->VDc);
    const double& VDp = *(((Params_t*)params)->VDp);
    const double& VDn = *(((Params_t*)params)->VDn);
    const double& CL = *(((Params_t*)params)->CL);
    const double& FE = *(((Params_t*)params)->FE);
    const double& PC = *(((Params_t*)params)->PC);
    const double& DE = *(((Params_t*)params)->DE);

//    0 = absComp1
//    1 = absComp2
//    2 = perComp
//    3 = bloodComp
//    4 = lungComp

    vecwrite[0] = -ka * vecread[0];
    vecwrite[1] = (ka * (vecread[0] - vecread[1])) - (FE*vecread[1]);
    vecwrite[2] = (Qp * (vecread[3]/VDc - vecread[2]/VDp)) - (CL*(vecread[2]/VDp));
    vecwrite[3] = (ka * vecread[1]) - (Qp * ((vecread[3]/VDc) - (vecread[2]/VDp))) - (Qn * ((PC*(vecread[3]/VDc)) - ((vecread[4]/VDn))));
    vecwrite[4] = (Qn * ((PC*(vecread[3]/VDc)) - (vecread[4]/VDn))) - (DE*(vecread[4]/VDn));
}

void BloodFuncTwoComp::operator()(const ODESolvers::ODEState& vecread, double, ODESolvers::Derivative& vecwrite, void* params) const
{
    const double& ka = *(((Params_t*)params)->ka);
    const double& Q = *(((Params_t*)params)->Q);
    const double& VDc = *(((Params_t*)params)->VDc);
    const double& VDp = *(((Params_t*)params)->VDp);
    const double& CL = *(((Params_t*)params)->CL);

//    0 = absComp1
//    1 = perComp
//    2 = bloodComp

    vecwrite[0] = -ka * vecread[0];
    vecwrite[1] = Q * ( vecread[2]/VDc - vecread[1]/VDp);
    vecwrite[2] = ka * vecread[0] - Q * ( vecread[2]/VDc - vecread[1]/VDp) - CL*vecread[2]/VDc;
}

