/*
* grvascular.h
*
*
* Created by ncilfone on 9/25/12.
*/


#ifndef GRVASCULAR_H
#define GRVASCULAR_H

#include "grgrid.h"
#include "gr.h"
#include "numericalMethods.h"

#include <sstream>

using std::valarray;

struct BloodFunc;
struct BloodFuncLung;
struct BloodFuncTwoComp;
enum PKModel {PKM_2_COMPARTMENT_1_ABSORPTION, PKM_2_COMPARTMENT_2_ABSORPTION};
class Vascular
{
private:
    double _bloodConcentrationIL10;         // Concentration in plasma/blood of IL-10
    double _bloodConcentrationTNF;          // Concentration in plasma/blood of TNF
    double _bloodConcentrationEtan;         // Concentration in plasma/blood of Etanercept
    double _bloodConcentrationInfliximab;          // Concentration in plasma/blood of Infliximab
    double _bloodConcentrationEtanTNF;         // Concentration in plasma/blood of Etanercept:TNF complex
    double _bloodConcentrationIfx1TNF;          // Concentration in plasma/blood of (Infliximab)1:TNF complex
    double _bloodConcentrationIfx2TNF;         // Concentration in plasma/blood of (Infliximab)2:TNF complex
    double _bloodConcentrationIfx3TNF;          // Concentration in plasma/blood of (Infliximab)3:TNF complex

    double _cutOffValue; // For setting drug amount in plasma ODE's to 0 to prevent flux errors when dosing of drug stops


    void updateBloodODE(double& bloodConcentration, double &absComp1Concentration, double &absComp2Concentration, double &perCompConcentration,
                        double ka, double Q, double VDc, double VDp, double CL, double MW, double dt, ODESolvers::ODEMethod odeSolver, std::valarray<double> &initvector);

    void updateBloodODE(double& bloodConcentration, double &absComp1Concentration, double &perCompConcentration,
                        double ka, double Q, double VDc, double VDp, double CL, double MW, double dt, ODESolvers::ODEMethod odeSolver, std::valarray<double> &initvector);

    void updateBlood1Compartment(double& bloodconc, const double clearance, const double dt);
    void addDose(const int time, double dosage, const int doseInterval, double& absComp1Concentration, int startTime);
    bool isDoseStartTime(const int time);
    bool isDoseStartTimeMod(const int time);
    void modulateBlood(double& bloodconc, double change, Scalar MW);
    void modulateBlood(double& bloodconc, double change, Scalar MW, Scalar volDist);
    void solveMolecule();
    double calculateFluxChangeDrugs(double localconc, double bloodconc,
                                    double dt, double permeability, Scalar MW,
                                    double permCoeff, const Pos& pos);
    double calculateFluxChange(double localconc, double bloodconc, double dt, double permeability, Scalar MW, Scalar volDist);

public:
    Vascular();
    virtual ~Vascular();
    void solveVascularSources(GrSimulation& grSim, GrGrid& grid, double dt,
                              const int time, int DiffStep,
                              const double timestepDiffusion,
                              ODESolvers::ODEMethod odeSolver);
    void solveVascularSourcesMod(GrGrid& grid, double dt, const int time,
                                 int DiffStep, const double timestepDiffusion);
    void solveVascularSourcesO2(GrGrid& grid, double dt);
    void updateState();
    double getAverageOfNeighbours(Scalar* grid, const Pos& pos, const Pos& dim, const Scalar MW);
    template<class Archive>
    void serialize(Archive& ar, const unsigned int version);
    double getBloodConcentrationEtan();
    double getbloodConcentrationInfliximab();
    double getBloodConcentrationEtanTNF();
    double getbloodConcentrationIfx1TNF();
    double getbloodConcentrationIfx2TNF();
    double getbloodConcentrationIfx3TNF();
    void writeValarrayFromMembers(valarray<double>& inputVector, double blComp, double &absComp1Concentration, double &absComp2Concentration, double &perCompConcentration);
    void writeValarrayFromMembers(valarray<double>& inputVector, double blComp, double &absComp1Concentration, double &absComp2Concentration, double &perCompConcentration, double& lungCompConcentration);
    void writeValarrayFromMembers(valarray<double>& inputVector, double blComp, double &absComp1Concentration, double &perCompConcentration);
    void writeMembersFromValarray(valarray<double>& inputVector, double MW, double VDc, double &bloodConcentration, double &absComp1Concentration, double &absComp2Concentration, double &perCompConcentration);
    void writeMembersFromValarray(valarray<double>& inputVector, double MW, double VDc, double &bloodConcentration, double &absComp1Concentration, double &absComp2Concentration, double &perCompConcentration, double& lungCompConcentration);
    void writeMembersFromValarray(valarray<double>& inputVector, double MW, double VDc, double &bloodConcentration, double &absComp1Concentration, double &perCompConcentration);

    static std::auto_ptr<ODESolvers::Stepper> stepper; /// ODE Stepper for all agents (unless overridden)
    static std::auto_ptr<BloodFunc> deriv;              /// ODE Function for all agents (unless overridden)
    static std::auto_ptr<BloodFuncLung> derivLung;
    static std::auto_ptr<BloodFuncTwoComp> derivTwoComp;              /// ODE Function for all agents (unless overridden)

    virtual ODESolvers::Stepper* getStepper(ODESolvers::ODEMethod odeSolver, int size)
    {
      static std::auto_ptr<ODESolvers::Stepper> stepper;
      static ODESolvers::ODEMethod initMethod = odeSolver;
      if(stepper.get() == NULL || odeSolver != initMethod)
        {
          stepper.reset(ODESolvers::StepperFactory(odeSolver, size));
          initMethod = odeSolver;
        }
      return stepper.get();
    }
    static BloodFunc* buildDerivFunc();
    static BloodFuncLung* buildDerivFuncLung();
    static BloodFuncTwoComp* buildDerivFuncTwoComp();

    virtual BloodFunc* getDerivFunc()
    {
        static std::auto_ptr<BloodFunc> deriv;              /// ODE Function for all agents (unless overridden)
        if(deriv.get() == NULL)
          {
            deriv.reset(buildDerivFunc());
          }
        return deriv.get();
    }

    virtual BloodFuncLung* getDerivFuncLung()
    {
        static std::auto_ptr<BloodFuncLung> derivLung;              /// ODE Function for all agents (unless overridden)
        if(derivLung.get() == NULL)
          {
            derivLung.reset(buildDerivFuncLung());
          }
        return derivLung.get();
    }

    virtual BloodFuncTwoComp* getDerivFuncTwoComp()
    {
        static std::auto_ptr<BloodFuncTwoComp> derivTwoComp;              /// ODE Function for all agents (unless overridden)
        if(derivTwoComp.get() == NULL)
          {
            derivTwoComp.reset(buildDerivFuncTwoComp());
          }
        return derivTwoComp.get();
    }

    Vascular* clone();
};

inline void Vascular::addDose(const int time, double dosage, const int doseInterval, double& absComp1Concentration, int startTime)
{
    int timeSinceStart = time - startTime * AGENT_TIME_STEP;
    // dosage is divided by two for all abx and immunomod because it is called twice for each timestep in grsimulation/solveVascularSourcesMod
    double halfdose = dosage/2;
    absComp1Concentration = (((timeSinceStart + doseInterval*AGENT_TIME_STEP) % (doseInterval*AGENT_TIME_STEP)) == 0) ? absComp1Concentration + halfdose : absComp1Concentration;
}

inline bool Vascular::isDoseStartTime(const int realTimeSeconds)
{
    return realTimeSeconds >= _PARAM(Drugs_dosageStartTime) * AGENT_TIME_STEP;
}

inline bool Vascular::isDoseStartTimeMod(const int realTimeSeconds)
{
    return realTimeSeconds >= ((_PARAM(Mod_dosageStartTime) * AGENT_TIME_STEP));
}

inline double Vascular::calculateFluxChangeDrugs(double localconc,
                            double bloodconc, double dt, double permeability,
                            Scalar MW, double permCoeff, const Pos& pos)
{
    const double vascularSurfaceArea = GridSpec::MICROCMPT_AREA_2D_CENTIMETERS_SQUARED;
    double change = permeability * vascularSurfaceArea * (permCoeff*bloodconc - localconc) / 1000.0 * dt * MW / VOL;

    if(fabs(change) > (fabs((permCoeff*bloodconc-localconc))*MW))
    {
        std::stringstream ss;
        ss << "Fatal error in Vascular::calculateFluxChangeDrugs: fabs(change) > fabs((permCoeff*bloodconc-localconc))*MW" << std::endl;
        ss << "pos: " << pos << std::endl;
		ss << "localconc: " << localconc << std::endl;
		ss << "bloodconc: " << bloodconc << std::endl;
		ss << "dt: " << dt << std::endl;
		ss << "permeability: " << permeability << std::endl;
		ss << "MW: " << MW << std::endl;
		ss << "permCoeff: " << permCoeff << std::endl;
		ss << "vascularSurfaceArea: " << vascularSurfaceArea << std::endl;
		ss << "change: " << change << std::endl;
		ss << "fabs(change): " << fabs(change) << std::endl;
		ss << "bloodconc-localconc: " << bloodconc-localconc << std::endl;
        ss << "fabs((permCoeff*bloodconc-localconc)): " << fabs((permCoeff*bloodconc-localconc)) << std::endl;
        ss << "fabs((permCoeff*bloodconc-localconc))*MW: " << fabs((permCoeff*bloodconc-localconc))*MW << std::endl;
        ss << "permeability * vascularSurfaceArea: " << permeability * vascularSurfaceArea << endl;
        throw std::runtime_error(ss.str());
    }
    return change;   // mg/L
}

inline double Vascular::calculateFluxChange(double localconc, double bloodconc, double dt, double permeability, Scalar MW, Scalar volDist)
{
    // LocalConc = Moles
    // BloodConc = ug/kg
    // Calculate the diff in correct units
    double diff = (bloodconc/(volDist * 1e6 * MW * 1000)) - (localconc/(VOL*1000)); // mol/cm^3

    double change = permeability * GridSpec::MICROCMPT_AREA_2D_CENTIMETERS_SQUARED * diff * dt; // mol/cm^3
	if(fabs(change) > fabs((bloodconc-localconc))*MW)
	{
		std::stringstream ss;
		ss << "Fatal error: fabs(change) > fabs((bloodconc-localconc))*MW" << std::endl;
		ss << "localconc: " << localconc << std::endl;
		ss << "bloodconc: " << bloodconc << std::endl;
		ss << "dt: " << dt << std::endl;
		ss << "permeability: " << permeability << std::endl;
		ss << "MW: " << MW << std::endl;
		ss << "volDist: " << volDist << std::endl;
		ss << "diff: " << diff << std::endl;
		ss << "change: " << change << std::endl;
		ss << "fabs(change): " << fabs(change) << std::endl;
		ss << "bloodconc-localconc: " << bloodconc-localconc << std::endl;
		ss << "fabs((bloodconc-localconc)): " << fabs((bloodconc-localconc)) << std::endl;
		ss << "fabs((bloodconc-localconc))*MW: " << fabs((bloodconc-localconc))*MW << std::endl;
		throw std::runtime_error(ss.str());
	}
    return change;   // mol/cm^3
}

inline void Vascular::modulateBlood(double &bloodconc, double change, Scalar MW)
{
    double change_molecules = change * VOL * NAV / MW;
    bloodconc = ((bloodconc * NAV * _PARAM(Drugs_bloodVolume)) - change_molecules)/(NAV*_PARAM(Drugs_bloodVolume));
}

inline void Vascular::modulateBlood(double &bloodconc, double change, Scalar MW, Scalar volDist)
{
    double change_units = change * (volDist * 1e6 * MW * 1000);
    bloodconc = (bloodconc - change_units);
//    std::cout << "modBlood: " << change_units << std::endl;
}

inline double Vascular::getBloodConcentrationEtan()
{
    return _bloodConcentrationEtan;
    cerr << "from Vascular::getBloodConcentrationEtan(), " << "Blood Concentration: " << _bloodConcentrationEtan << endl;
}

inline double Vascular::getbloodConcentrationInfliximab()
{
    return _bloodConcentrationInfliximab;
}
inline double Vascular::getBloodConcentrationEtanTNF()
{
    return _bloodConcentrationEtanTNF;
}

inline double Vascular::getbloodConcentrationIfx1TNF()
{
    return _bloodConcentrationIfx1TNF;
}
inline double Vascular::getbloodConcentrationIfx2TNF()
{
    return _bloodConcentrationIfx2TNF;
}

inline double Vascular::getbloodConcentrationIfx3TNF()
{
    return _bloodConcentrationIfx3TNF;
}

struct BloodFunc : ODESolvers::DerivativeFunc
{
    struct Params_t
    {
        double* ka;
        double* Q;
        double* VDc;
        double* VDp;
        double* CL;
    };
    BloodFunc() : DerivativeFunc() {}
    virtual ~BloodFunc() {}
    /*virtual*/ void operator()(const ODESolvers::ODEState& vecread, double t, ODESolvers::Derivative& vecwrite, void* params) const;

};

struct BloodFuncLung:ODESolvers::DerivativeFunc
{
    struct Params_t
    {
        double* ka;
        double* Qp;
        double* Qn;
        double* VDc;
        double* VDp;
        double* VDn;
        double* CL;
        double* FE;
        double* PC;
        double* DE;
    };
    BloodFuncLung() : DerivativeFunc() {}
    virtual ~BloodFuncLung() {}
    /*virtual*/ void operator()(const ODESolvers::ODEState& vecread, double t, ODESolvers::Derivative& vecwrite, void* params) const;
};

struct BloodFuncTwoComp : ODESolvers::DerivativeFunc
{
    struct Params_t
    {
        double* ka;
        double* Q;
        double* VDc;
        double* VDp;
        double* CL;
    };
    BloodFuncTwoComp() : DerivativeFunc() {}
    virtual ~BloodFuncTwoComp() {}
    /*virtual*/ void operator()(const ODESolvers::ODEState& vecread, double t, ODESolvers::Derivative& vecwrite, void* params) const;

};

inline BloodFunc* Vascular::buildDerivFunc()
{
  BloodFunc* d = new BloodFunc();
  return d;
}

inline BloodFuncLung* Vascular::buildDerivFuncLung()
{
    BloodFuncLung* d = new BloodFuncLung();
    return d;
}

inline BloodFuncTwoComp* Vascular::buildDerivFuncTwoComp()
{
  BloodFuncTwoComp* d = new BloodFuncTwoComp();
  return d;
}

template<class Archive>
void Vascular::serialize(Archive& ar, const unsigned int /*version*/)
{
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationIL10);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationTNF);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationEtan);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationInfliximab);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationEtanTNF);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationIfx1TNF);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationIfx2TNF);
    ar & BOOST_SERIALIZATION_NVP(_bloodConcentrationIfx3TNF);

    ar & BOOST_SERIALIZATION_NVP(_cutOffValue);
}

#endif /* GRVASCULAR_H */
